# NOTE: this is copied from Tom Augspurger' stac-table repository. We had to adopt some
# change some parts and installing the package on remote machines from gh was a bit tedious.
# Try to keep this code in sync with the original and remove this note when the package
# can be installed. Also please acknowledge the original author.
"""
Generate STAC Collections for tabular datasets.
"""

import copy
import enum
from typing import TypeVar

import dask

# NOTE: until query planning is enabled in Dask GeoPandas
dask.config.set({"dataframe.query-planning": False})

import dask_geopandas
import fsspec
import numpy as np
import pandas as pd
import pyarrow as pa
import pyproj
import pystac
import shapely.geometry
from shapely.ops import transform

T = TypeVar("T", pystac.Collection, pystac.Item)
SCHEMA_URI = "https://stac-extensions.github.io/table/v1.2.0/schema.json"
# https://issues.apache.org/jira/browse/PARQUET-1889: parquet doesn't officially
# have a type yet.
# NOTE: follow discussion at: https://github.com/opengeospatial/geoparquet/issues/115
PARQUET_MEDIA_TYPE = "application/vnd.apache.parquet"


class InferDatetimeOptions(str, enum.Enum):
    no = "no"
    midpoint = "midpoint"
    unique = "unique"
    range = "range"


def generate(
    uri: str,
    template,
    infer_bbox=None,
    infer_geometry=False,
    datetime_column=None,
    infer_datetime=InferDatetimeOptions.no,
    count_rows=True,
    asset_key="data",
    asset_extra_fields=None,
    proj=True,
    storage_options=None,
    validate=True,
) -> T:
    """
    Generate a STAC Item from a Parquet Dataset.

    Parameters
    ----------
    uri : str
        The fsspec-compatible URI pointing to the input table to generate a
        STAC item for.
    template : pystac.Item
        The template item. This will be cloned and new data will be filled in.
    infer_bbox : str, optional
        The column name to use setting the Item's bounding box.

        .. note::

           If the dataset doesn't provide spatial partitions, this will
           require computation.

    infer_geometry: bool, optional
        Whether to fill the item's `geometry` field with the union of the
        geometries in the `infer_bbox` column.

    datetime_column: str, optional
        The column name to use when setting the Item's `datetime` or
        `start_datetime` and `end_datetime` properties. The method used is
        determined by `infer_datetime`.

    infer_datetime: str, optional.
        The method used to find a datetime from the values in `datetime_column`.
        Use the options in the `InferDatetimeOptions` enum.

        - no : do not infer a datetime
        - midpoint : Set `datetime` to the midpoint of the highest and lowest values.
        - unique : Set `datetime` to the unique value. Raises if more than one
          unique value is found.
        - range : Set `start_datetime` and `end_datetime` to the minimum and
          maximum values.

    count_rows : bool, default True
        Whether to add the row count to `table:row_count`.

    asset_key : str, default "data"
        The asset key to use for the parquet dataset. The href will be the ``uri`` and
        the roles will be ``["data"]``.

    asset_extra_fields : dict, optional
        Additional fields to set in the asset's ``extra_fields``.

    proj : bool or dict, default True
        Whether to extract projection information from the dataset and store it
        using the `projection` extension.

        By default, just `proj:crs` is extracted. If `infer_bbox` or `infer_geometry`
        are specified, those will be set as well.

        Alternatively, provide a dict of values to include.

    storage_options: mapping, optional
        A dictionary of keywords to provide to :meth:`fsspec.get_fs_token_paths`
        when creating an fsspec filesystem with a str ``ds``.

    validate : bool, default True
        Whether to validate the returned pystac.Item.

    Returns
    -------
    pystac.Item
        The updated pystac.Item with the following fields set

        * stac_extensions : added `table` extension
        * table:columns

    Examples
    --------

    This example generates a STAC item based on the "naturalearth_lowres" datset
    from geopandas. There's a bit of setup.

    >>> import datetime, geopandas, pystac, stac_table
    >>> gdf = geopandas.read_file(geopandas.datasets.get_path("naturalearth_lowres"))
    >>> gdf.to_parquet("data.parquet")

    Now we can create the item.

    >>> # Create the template Item
    >>> item = pystac.Item(
    ...     "naturalearth_lowres",
    ...     geometry=None,
    ...     bbox=None,
    ...     datetime=datetime.datetime(2021, 1, 1),
    ...     properties={},
    ... )
    >>> result = stac_table.generate("data.parquet", item)
    >>> result
    <Item id=naturalearth_lowres>
    """
    template = copy.deepcopy(template)

    data = None
    storage_options = storage_options or {}
    # data = dask_geopandas.read_parquet(
    #     ds, storage_options=storage_options
    # )
    ds = parquet_dataset_from_url(uri, storage_options)

    if (
        infer_bbox
        or infer_geometry
        or infer_datetime != InferDatetimeOptions.no
        or proj is True
    ):
        # NOTE: gather spatial partitions has been temporarily disabled
        data = dask_geopandas.read_parquet(
            uri, storage_options=storage_options, gather_spatial_partitions=False
        )
        data.calculate_spatial_partitions()
    # #     # TODO: this doesn't actually work
    #     data = dask_geopandas.read_parquet(
    #         ds.files, filesystem=ds.filesystem,
    #     )

    columns = get_columns(ds.schema)
    template.properties["table:columns"] = columns

    if proj is True:
        proj = get_proj(data)
    proj = proj or {}

    # TODO: Add schema when published
    if SCHEMA_URI not in template.stac_extensions:
        template.stac_extensions.append(SCHEMA_URI)
    if proj and pystac.extensions.projection.SCHEMA_URI not in template.stac_extensions:
        template.stac_extensions.append(pystac.extensions.projection.SCHEMA_URI)

    extra_proj = {}
    if infer_bbox:
        src_crs = data.spatial_partitions.crs.to_epsg()
        tf = pyproj.Transformer.from_crs(src_crs, 4326, always_xy=True)

        bbox = data.spatial_partitions.unary_union.bounds
        # NOTE: bbox of unary union will be stored under proj extension as projected
        extra_proj["proj:bbox"] = bbox

        # NOTE: bbox will be stored in pystsac.Item.bbox in EPSG:4326
        bbox = transform(tf.transform, shapely.geometry.box(*bbox))
        template.bbox = bbox.bounds

    if infer_geometry:
        # NOTE: geom  under proj extension as projected
        geometry = data.unary_union.compute()
        extra_proj["proj:geometry"] = shapely.geometry.mapping(geometry)

        # NOTE: geometry will be stored in pystsac.Item.geometry in EPSG:4326
        src_crs = data.spatial_partitions.crs.to_epsg()
        tf = pyproj.Transformer.from_crs(src_crs, 4326, always_xy=True)
        geometry = transform(tf.transform, geometry)
        template.geometry = shapely.geometry.mapping(geometry)

    if infer_bbox and template.geometry is None:
        # If bbox is set then geometry must be set as well.
        template.geometry = shapely.geometry.mapping(
            shapely.geometry.box(*template.bbox)
        )

    if infer_geometry and template.bbox is None:
        template.bbox = shapely.geometry.shape(template.geometry).bounds

    if proj or extra_proj:
        template.properties.update(**extra_proj, **proj)

    if infer_datetime != InferDatetimeOptions.no and datetime_column is None:
        msg = "Must specify 'datetime_column' when 'infer_datetime != no'."
        raise ValueError(msg)

    if infer_datetime == InferDatetimeOptions.midpoint:
        values = dask.compute(data[datetime_column].min(), data[datetime_column].max())
        template.properties["datetime"] = pd.Series(values).mean().to_pydatetime()

    if infer_datetime == InferDatetimeOptions.unique:
        values = data[datetime_column].unique().compute()
        n = len(values)
        if n > 1:
            msg = f"infer_datetime='unique', but {n} unique values found."
            raise ValueError(msg)
        template.properties["datetime"] = values[0].to_pydatetime()

    if infer_datetime == InferDatetimeOptions.range:
        values = dask.compute(data[datetime_column].min(), data[datetime_column].max())
        values = np.array(pd.Series(values).dt.to_pydatetime())
        template.properties["start_datetime"] = values[0].isoformat() + "Z"
        template.properties["end_datetime"] = values[1].isoformat() + "Z"

    if count_rows:
        template.properties["table:row_count"] = sum(
            x.count_rows() for x in ds.fragments
        )

    if asset_key:
        asset = pystac.asset.Asset(
            # NOTE: consider using the https protocol; makes it easier for user to download
            # to_https_url(items[0].assets["data"].href, storage_options={"account_name": "coclico"})
            uri,
            title="Dataset root",
            media_type=PARQUET_MEDIA_TYPE,
            roles=["data"],
            # extra_fields={"table:storage_options": asset_extra_fields},
            extra_fields=asset_extra_fields,
        )
        template.add_asset(asset_key, asset)

    if validate:
        template.validate()

    return template


def get_proj(ds):
    """
    Read projection information from the dataset.
    """
    # Use geopandas to get the proj info
    proj = {}
    maybe_crs = ds.geometry.crs
    if maybe_crs:
        maybe_epsg = ds.geometry.crs.to_epsg()
        if maybe_epsg:
            proj["proj:epsg"] = maybe_epsg
        else:
            proj["proj:wkt2"] = ds.geometry.crs.to_wkt()

    return proj


def get_columns(schema: pa.Schema, prefix: str = "") -> list:
    columns = []
    for field in schema:
        if field.name == "__null_dask_index__":
            continue
        # Check if the field is nested
        if pa.types.is_struct(field.type):
            # For nested fields, recurse into the structure
            nested_columns = get_columns(field.type, prefix=field.name + ".")
            columns.extend(nested_columns)
        elif field.name == "geometry":
            column = {"name": "geometry", "type": "WKB"}
            columns.append(column)

        else:
            # Handle non-nested fields
            column = {"name": prefix + field.name, "type": str(field.type).lower()}
            if field.metadata is not None:
                column["metadata"] = field.metadata
            columns.append(column)
    return columns


def parquet_dataset_from_url(url: str, storage_options):
    fs, _, _ = fsspec.get_fs_token_paths(url, storage_options=storage_options)
    pa_fs = pa.fs.PyFileSystem(pa.fs.FSSpecHandler(fs))
    url2 = url.split("://", 1)[-1]  # pyarrow doesn't auto-strip the prefix.
    # NOTE: use_legacy_dataset will be deprecated, so test if it works without it.
    # ds = pa.parquet.ParquetDataset(url2, filesystem=pa_fs, use_legacy_dataset=False)
    ds = pa.parquet.ParquetDataset(url2, filesystem=pa_fs)
    return ds



import os
from typing import Literal

import duckdb
import geopandas as gpd
import pystac
import shapely
from shapely.wkb import loads as wkb_loads

from coastpy.io.utils import read_items_extent


class BaseQueryEngine:
    """
    Base class for querying geospatial data using DuckDB.

    Attributes:
        storage_backend (Literal["azure", "aws"]): The cloud storage backend.
        con (duckdb.DuckDBPyConnection): DuckDB connection object.
    """

    def __init__(self, storage_backend: Literal["azure", "aws"] = "azure") -> None:
        self.storage_backend = storage_backend
        self.con = duckdb.connect(database=":memory:", read_only=False)
        self._initialize_spatial_extension()
        self._configure_storage_backend()

    def _initialize_spatial_extension(self) -> None:
        """Initializes the spatial extension in DuckDB."""
        self.con.execute("INSTALL spatial;")
        self.con.execute("LOAD spatial;")

    def _configure_storage_backend(self) -> None:
        """Configures the storage backend for DuckDB."""
        if self.storage_backend == "azure":
            self.con.execute("INSTALL azure;")
            self.con.execute("LOAD azure;")

        elif self.storage_backend == "aws":
            self.con.execute("INSTALL httpfs;")
            self.con.execute("LOAD httpfs;")
            self.con.execute(f"SET s3_region = '{os.getenv('AWS_REGION')}';")
            self.con.execute(
                f"SET s3_access_key_id = '{os.getenv('AWS_ACCESS_KEY_ID')}';"
            )
            self.con.execute(
                f"SET s3_secret_access_key = '{os.getenv('AWS_SECRET_ACCESS_KEY')}';"
            )

    def _get_token(self) -> str:
        """
        Retrieve the storage SAS token based on the storage backend.

        Returns:
            str: The SAS token if the storage backend is Azure and the token is available, otherwise an empty string.
        """
        if self.storage_backend == "azure":
            sas_token = os.getenv("AZURE_STORAGE_SAS_TOKEN")
            return sas_token if sas_token else ""
        return ""

    def execute_query(self, query: str) -> gpd.GeoDataFrame:
        """
        Executes a SQL query and returns the result as a GeoDataFrame.

        Args:
            query (str): The SQL query to execute.

        Returns:
            gpd.GeoDataFrame: The query result as a GeoDataFrame.
        """
        df = self.con.execute(query).fetchdf()
        if not df.empty:
            df["geometry"] = df.geometry.map(lambda b: wkb_loads(bytes(b)))
            return gpd.GeoDataFrame(df, crs="EPSG:4326")
        return gpd.GeoDataFrame()


class STACQueryEngine(BaseQueryEngine):
    """
    A query engine optimized by STAC for querying geospatial data.

    Inherits from BaseQueryEngine.

    Attributes:
        quadtiles (gpd.GeoDataFrame): GeoDataFrame of quadtiles from the STAC collection.
        proj_epsg (int): EPSG code of the projection used in the quadtiles.
    """

    def __init__(
        self,
        stac_collection: pystac.Collection,
        storage_backend: Literal["azure", "aws"] = "azure",
        columns: list[str] | None = None,
    ) -> None:
        super().__init__(storage_backend=storage_backend)
        self.extents = read_items_extent(
            stac_collection, columns=["geometry", "assets", "proj:epsg", "href"]
        )
        self.proj_epsg = self.extents["proj:epsg"].unique().item()

        if columns is None or not columns:
            self.columns = ["*"]
        else:
            self.columns = columns

    def get_data_within_bbox(self, minx: float, miny: float, maxx: float, maxy: float):
        """
        Retrieve data within the specified bounding box.

        Args:
            minx (float): Minimum X coordinate of the bounding box.
            miny (float): Minimum Y coordinate of the bounding box.
            maxx (float): Maximum X coordinate of the bounding box.
            maxy (float): Maximum Y coordinate of the bounding box.

        Returns:
            pd.DataFrame: The queried data.
        """
        bbox = shapely.box(minx, miny, maxx, maxy)
        bbox_gdf = gpd.GeoDataFrame(geometry=[bbox], crs="EPSG:4326")

        # Perform spatial join to get all overlapping HREFs
        overlapping_hrefs = gpd.sjoin(self.extents, bbox_gdf).href.tolist()

        # Sign each HREF with the SAS token if the storage backend is Azure
        if self.storage_backend == "azure":
            signed_hrefs = []
            for href in overlapping_hrefs:
                signed_href = href.replace(
                    "az://", "https://coclico.blob.core.windows.net/"
                )
                signed_href = signed_href + "?" + self._get_token()
                signed_hrefs.append(signed_href)
        else:
            signed_hrefs = overlapping_hrefs

        # Join the hrefs into a single string
        hrefs_str = ", ".join(f'"{href}"' for href in signed_hrefs)

        # Construct and execute the query
        columns_str = ", ".join(self.columns)
        query = f"""
        SELECT {columns_str}
        FROM read_parquet([{hrefs_str}])
        WHERE
            bbox.xmin <= {maxx} AND
            bbox.ymin <= {maxy} AND
            bbox.xmax >= {minx} AND
            bbox.ymax >= {miny};
        """
        return self.execute_query(query)


class HREFQueryEngine(BaseQueryEngine):
    """
    A query engine for directly querying geospatial data using an HREF.

    Inherits from BaseQueryEngine.

    Attributes:
        href (str): The direct link (HREF) to the geospatial data.
    """

    def __init__(
        self, href: str, storage_backend: Literal["azure", "aws"] = "azure"
    ) -> None:
        super().__init__(storage_backend=storage_backend)
        self.href = href

    def get_data_within_bbox(
        self, minx: float, miny: float, maxx: float, maxy: float
    ) -> gpd.GeoDataFrame:
        """
        Queries data within a specified bounding box from a direct HREF.

        Args:
            minx (float): Minimum X coordinate (longitude) of the bounding box.
            miny (float): Minimum Y coordinate (latitude) of the bounding box.
            maxx (float): Maximum X coordinate (longitude) of the bounding box.
            maxy (float): Maximum Y coordinate (latitude) of the bounding box.

        Returns:
            gpd.GeoDataFrame: Queried data within the bounding box as a GeoDataFrame.
        """
        query = f"""
        SELECT *
        FROM read_parquet('{self.href}', hive_partitioning = false)
        WHERE
            bbox.xmin <= {maxx} AND
            bbox.ymin <= {maxy} AND
            bbox.xmax >= {minx} AND
            bbox.ymax >= {miny};
        """
        return self.execute_query(query)



from .geoms import create_offset_rectangle
from .quadtiles import make_mercantiles
from .quadtiles_utils import add_geo_columns

__all__ = ["add_geo_columns", "make_mercantiles", "create_offset_rectangle"]



import dataclasses
import datetime
import os
import pathlib
import re
from typing import Any

import fsspec
import pystac
import stac_geoparquet
from dotenv import load_dotenv
from pystac.extensions.item_assets import ItemAssetsExtension
from pystac.extensions.scientific import ScientificExtension
from pystac.extensions.version import VersionExtension
from pystac.provider import ProviderRole
from pystac.stac_io import DefaultStacIO

from coastpy.libs import stac_table
from coastpy.stac import ParquetLayout

# Load the environment variables from the .env file
load_dotenv()

# Get the SAS token and storage account name from environment variables
sas_token = os.getenv("AZURE_STORAGE_SAS_TOKEN")
storage_account_name = "coclico"
storage_options = {"account_name": storage_account_name, "credential": sas_token}

# NOTE:
TEST_RELEASE = True

# Container and URI configuration
CONTAINER_NAME = "gcts"
RELEASE_DATE = "2024-08-02"
PREFIX = f"release/{RELEASE_DATE}"
CONTAINER_URI = f"az://{CONTAINER_NAME}/{PREFIX}"
PARQUET_MEDIA_TYPE = "application/vnd.apache.parquet"
LICENSE = "CC-BY-4.0"

# Collection information
COLLECTION_ID = "gcts"
COLLECTION_TITLE = "Global Coastal Transect System (GCTS)"

# Transect and zoom configuration
TRANSECT_LENGTH = 2000
ZOOM = 9

DESCRIPTION = """
Cross-shore coastal transects are essential to coastal monitoring, offering a consistent
reference line to measure coastal change, while providing a robust foundation to map
coastal characteristics and derive coastal statistics thereof. The Global Coastal Transect
System consists of more than 11 million cross-shore coastal transects uniformly spaced at
100-m intervals alongshore, for all OpenStreetMap coastlines that are longer than 5 kilometers.
The dataset is more extensively described Calkoen et al., 2024. "Enabling Coastal Analytics
at Planetary Scale" available [here](https://doi.org/10.1016/j.envsoft.2024.106257).
"""
# Asset details
ASSET_TITLE = "GCTS"
ASSET_DESCRIPTION = f"Parquet dataset with coastal transects ({TRANSECT_LENGTH} m) at 100 m alongshore resolution for this region."

# GeoParquet STAC items
if TEST_RELEASE:
    GEOPARQUET_STAC_ITEMS_HREF = f"az://items-test/{COLLECTION_ID}.parquet"
else:
    GEOPARQUET_STAC_ITEMS_HREF = f"az://items/{COLLECTION_ID}.parquet"

COLUMN_DESCRIPTIONS = [
    {
        "name": "transect_id",
        "type": "string",
        "description": "A unique identifier for each transect, constructed from three key components: the 'coastline_id', 'segment_id', and 'interpolated_distance'. The 'coastline_id' corresponds to the FID in OpenStreetMap (OSM) and is prefixed with 'cl'. The 'segment_id' indicates the segment of the OSM coastline split by a UTM grid, prefixed with 's'. The 'interpolated_distance' represents the distance from the starting point of the coastline to the transect, interpolated along the segment, and is prefixed with 'tr'. The complete structure is 'cl[coastline_id]s[segment_id]tr[interpolated_distance]', exemplified by 'cl32946s04tr08168547'. This composition ensures each transect name is a distinct and informative representation of its geographical and spatial attributes.",
    },
    {
        "name": "lon",
        "type": "float",
        "description": "Longitude of the transect origin.",
    },
    {
        "name": "lat",
        "type": "float",
        "description": "Latitude of the transect origin.",
    },
    {
        "name": "bearing",
        "type": "float",
        "description": "North bearing of the transect from the landward side in degrees, with the north as reference.",
    },
    {
        "name": "geometry",
        "type": "byte_array",
        "description": "Well-Known Binary (WKB) representation of the transect as a linestring geometry.",
    },
    {
        "name": "osm_coastline_is_closed",
        "type": "bool",
        "description": "Indicates whether the source OpenStreetMap (OSM) coastline, from which the transects were derived, forms a closed loop. A value of 'true' suggests that the coastline represents an enclosed area, such as an island.",
    },
    {
        "name": "osm_coastline_length",
        "type": "int32",
        "description": "Represents the total length of the source OpenStreetMap (OSM) coastline, that is summed across various UTM regions. It reflects the aggregate length of the original coastline from which the transects are derived.",
    },
    {
        "name": "utm_epsg",
        "type": "int32",
        "description": "EPSG code representing the UTM Coordinate Reference System for the transect.",
    },
    {
        "name": "bbox",
        "type": "struct<minx: double, miny: double, maxx: double, maxy: double>",
        "description": "Bounding box of the transect geometry, given by minimum and maximum coordinates in x (longitude) and y (latitude).",
    },
    {
        "name": "quadkey",
        "type": "string",
        "description": "QuadKey corresponding to the transect origin location at zoom 12, following the Bing Maps Tile System for spatial indexing.",
    },
    {
        "name": "continent",
        "type": "string",
        "description": "Name of the continent in which the transect is located.",
    },
    {
        "name": "country",
        "type": "string",
        "description": "ISO alpha-2 country code for the country in which the transect is located. The country data are extracted from Overture Maps (divisions).",
    },
    {
        "name": "common_country_name",
        "type": "string",
        "description": "Common country name (EN) in which the transect is located. The country data are extracted from Overture Maps (divisions).",
    },
    {
        "name": "common_region_name",
        "type": "string",
        "description": "Common region name (EN) in which the transect is located. The regions are extracted from Overture Maps (divisions).",
    },
]


ASSET_EXTRA_FIELDS = {
    "table:storage_options": {"account_name": "coclico"},
    "table:columns": COLUMN_DESCRIPTIONS,
}


@dataclasses.dataclass
class PathParts:
    """
    Parses a path into its component parts, supporting variations with and without hive partitioning,
    and with and without geographical bounds.

    Attributes:
        path (str): The full path to parse.
        container (str | None): The storage container or bucket name.
        prefix (str | None): The path prefix between the container and the file name.
        name (str | None): The file name including its extension.
        stac_item_id (str | None): The identifier used for STAC, which is the file name without the .parquet extension.
    """

    path: str
    container: str | None = None
    prefix: str | None = None
    name: str | None = None
    stac_item_id: str | None = None

    def __post_init__(self) -> None:
        # Strip any protocol pattern like "az://"
        stripped_path = re.sub(r"^\w+://", "", self.path)
        split = stripped_path.rstrip("/").split("/")

        # Extract the container which is the first part of the path
        self.container = split[0]

        # The file name is the last element in the split path
        self.name = split[-1]

        # The prefix is everything between the container and the filename
        self.prefix = "/".join(split[1:-1]) if len(split) > 2 else None

        # stac_item_id is the filename without the .parquet extension
        self.stac_item_id = self.name.replace(".parquet", "")


def create_collection(
    description: str | None = None, extra_fields: dict[str, Any] | None = None
) -> pystac.Collection:
    providers = [
        pystac.Provider(
            name="Deltares",
            roles=[
                ProviderRole.PRODUCER,
                ProviderRole.PROCESSOR,
                ProviderRole.HOST,
                ProviderRole.LICENSOR,
            ],
            url="https://deltares.nl",
        ),
    ]

    start_datetime = datetime.datetime.strptime(RELEASE_DATE, "%Y-%m-%d")

    extent = pystac.Extent(
        pystac.SpatialExtent([[-180.0, 90.0, 180.0, -90.0]]),
        pystac.TemporalExtent([[start_datetime, None]]),
    )

    links = [
        pystac.Link(
            pystac.RelType.LICENSE,
            target="https://creativecommons.org/licenses/by/4.0/",
            media_type="text/html",
            title="CC BY 4.0 ",
        ),
    ]

    keywords = [
        "Coastal analytics",
        "Cloud technology",
        "Coastal change",
        "Coastal monitoring",
        "Satellite-derived shorelines",
        "Low elevation coastal zone",
        "Data management",
        "Transects",
        "GCTS",
        "Deltares",
        "CoCliCo",
        "GeoParquet",
    ]
    if description is None:
        description = DESCRIPTION

    collection = pystac.Collection(
        id=COLLECTION_ID,
        title=COLLECTION_TITLE,
        description=description,
        license=LICENSE,
        providers=providers,
        extent=extent,
        catalog_type=pystac.CatalogType.RELATIVE_PUBLISHED,
    )

    collection.add_asset(
        "thumbnail",
        pystac.Asset(
            "https://coclico.blob.core.windows.net/assets/thumbnails/gcts-thumbnail.jpeg",
            title="Thumbnail",
            media_type=pystac.MediaType.JPEG,
        ),
    )
    collection.links = links
    collection.keywords = keywords

    ItemAssetsExtension.add_to(collection)

    collection.extra_fields["item_assets"] = {
        "data": {
            "title": ASSET_TITLE,
            "description": ASSET_DESCRIPTION,
            "roles": ["data"],
            "type": stac_table.PARQUET_MEDIA_TYPE,
            **ASSET_EXTRA_FIELDS,
        }
    }

    if extra_fields:
        collection.extra_fields.update(extra_fields)

    ScientificExtension.add_to(collection)
    # TODO: revise citation upon publication

    CITATION = """
    Floris Reinier Calkoen, Arjen Pieter Luijendijk, Kilian Vos, Etiënne Kras, Fedor Baart,
    Enabling coastal analytics at planetary scale, Environmental Modelling & Software, 2024,
    106257, ISSN 1364-8152, https://doi.org/10.1016/j.envsoft.2024.106257.
    (https://www.sciencedirect.com/science/article/pii/S1364815224003189)
    """

    # NOTE: we could make a separate DOI for the transects and then link the paper in
    # sci:citations as a feature. However, for now I (Floris) prefer to use the same DOI.
    collection.extra_fields["sci:citation"] = CITATION
    # collection.extra_fields["sci:doi"] = "https://doi.org/10.1016/j.envsoft.2024.106257"

    collection.stac_extensions.append(stac_table.SCHEMA_URI)

    VersionExtension.add_to(collection)
    collection.extra_fields["version"] = RELEASE_DATE

    return collection


def create_item(
    asset_href: str,
    storage_options: dict[str, Any] | None = None,
    asset_extra_fields: dict[str, Any] | None = None,
) -> pystac.Item:
    """Create a STAC Item

    For

    Args:
        asset_href (str): The HREF pointing to an asset associated with the item

    Returns:
        Item: STAC Item object
    """

    parts = PathParts(asset_href)

    properties = {
        "title": ASSET_TITLE,
        "description": ASSET_DESCRIPTION,
    }

    dt = datetime.datetime.strptime(RELEASE_DATE, "%Y-%m-%d")
    # shape = shapely.box(*bbox)
    # geometry = shapely.geometry.mapping(shape)
    template = pystac.Item(
        id=parts.stac_item_id,
        properties=properties,
        geometry=None,
        bbox=None,
        datetime=dt,
        stac_extensions=[],
    )

    item = stac_table.generate(
        uri=asset_href,
        template=template,
        infer_bbox=True,
        infer_geometry=None,
        datetime_column=None,
        infer_datetime=stac_table.InferDatetimeOptions.no,
        count_rows=True,
        asset_key="data",
        asset_extra_fields=asset_extra_fields,
        proj=True,
        storage_options=storage_options,
        validate=False,
    )
    assert isinstance(item, pystac.Item)

    item.common_metadata.created = datetime.datetime.now(datetime.UTC)

    # add descriptions to item properties
    if "table:columns" in ASSET_EXTRA_FIELDS and "table:columns" in item.properties:
        source_lookup = {
            col["name"]: col for col in ASSET_EXTRA_FIELDS["table:columns"]
        }

    for target_col in item.properties["table:columns"]:
        source_col = source_lookup.get(target_col["name"])
        if source_col:
            target_col.setdefault("description", source_col.get("description"))

    # TODO: make configurable upstream
    item.assets["data"].title = ASSET_TITLE
    item.assets["data"].description = ASSET_DESCRIPTION

    return item


if __name__ == "__main__":
    storage_options = {"account_name": "coclico", "credential": sas_token}
    fs, token, [root] = fsspec.get_fs_token_paths(
        CONTAINER_URI, storage_options=storage_options
    )
    paths = fs.glob(CONTAINER_URI + "/**/*.parquet")
    uris = ["az://" + p for p in paths]

    STAC_DIR = pathlib.Path.home() / "dev" / "coclicodata" / "current"
    catalog = pystac.Catalog.from_file(str(STAC_DIR / "catalog.json"))

    stac_io = DefaultStacIO()
    layout = ParquetLayout()

    collection = create_collection(extra_fields={"container_uri": CONTAINER_URI})
    collection.validate_all()

    for uri in uris:
        item = create_item(uri, storage_options=storage_options)
        item.validate()
        collection.add_item(item)

    collection.update_extent_from_items()

    items = list(collection.get_all_items())
    items_as_json = [i.to_dict() for i in items]
    item_extents = stac_geoparquet.to_geodataframe(items_as_json)

    with fsspec.open(GEOPARQUET_STAC_ITEMS_HREF, mode="wb", **storage_options) as f:
        item_extents.to_parquet(f)

    collection.add_asset(
        "geoparquet-stac-items",
        pystac.Asset(
            GEOPARQUET_STAC_ITEMS_HREF,
            title="GeoParquet STAC items",
            description="Snapshot of the collection's STAC items exported to GeoParquet format.",
            media_type=PARQUET_MEDIA_TYPE,
            roles=["data"],
        ),
    )

    # TODO: there should be a cleaner method to remove the previous stac catalog and its items
    try:
        if catalog.get_child(collection.id):
            catalog.remove_child(collection.id)
            print(f"Removed child: {collection.id}.")
    except Exception:
        pass

    catalog.add_child(collection)

    collection.normalize_hrefs(str(STAC_DIR / collection.id), layout)

    collection.validate_all()

    catalog.save(
        catalog_type=pystac.CatalogType.SELF_CONTAINED,
        dest_href=str(STAC_DIR),
        stac_io=stac_io,
    )



from typing import Any

import geopandas as gpd
import pandas as pd
from shapely.geometry import LineString, Point


def linestring_to_coords(
    data: gpd.GeoDataFrame | gpd.GeoSeries | LineString,
    columns: list[str] | None = None,
) -> tuple[pd.Series, pd.Series] | gpd.GeoDataFrame:
    """
    Explode linestring geometry and return longitude and latitude coordinates
    as separate variables.

    Args:
        data (Union[gpd.GeoDataFrame, gpd.GeoSeries, LineString]): The input geometry.
        columns (List[str], optional): List containing names of longitude and latitude, defaults to ["lon", "lat"].

    Returns:
        Union[Tuple[pd.Series, pd.Series], gpd.GeoDataFrame]: If input is a GeoSeries,
            returns two Series with lon and lat values. If input is a GeoDataFrame,
            returns the dataframe with exploded coordinates.
    """

    if columns is None:
        columns = ["lon", "lat"]

    if isinstance(data, LineString):
        data = gpd.GeoSeries(data)

    exploded = data.geometry.apply(lambda x: x.coords).explode()

    if isinstance(data, gpd.GeoSeries):
        gs = gpd.GeoSeries(exploded.apply(Point))
        return gs.x, gs.y

    elif isinstance(data, gpd.GeoDataFrame):
        df = pd.DataFrame(exploded.tolist(), columns=columns, index=exploded.index)
        return data.join(df).reset_index().rename(columns={"index": "group"})

    msg = "data must be a GeoDataFrame, GeoSeries, or LineString"
    raise TypeError(msg)


def geojson_to_bbox(geometry: dict[str, Any]) -> list[float]:
    """
    Extract the bounding box from a geojson geometry.

    Args:
        geometry (Dict[str, Any]): GeoJSON geometry dictionary.

    Returns:
        List[float]: Bounding box of the geojson geometry, formatted according to:

    Note: https://tools.ietf.org/html/rfc7946#section-5

    Raises:
        ValueError: If the provided geometry type is not supported.
    """

    geom_type = geometry["type"]
    coords = geometry["coordinates"]

    if geom_type == "Point":
        lon, lat = coords
        return [lon, lat, lon, lat]

    lats, lons = [], []

    if geom_type in ["LineString", "MultiPoint"]:
        for lon, lat in coords:
            lons.append(lon)
            lats.append(lat)

    elif geom_type == "Polygon":
        for linear_ring in coords:
            for lon, lat in linear_ring:
                lons.append(lon)
                lats.append(lat)

    elif geom_type in ["MultiLineString", "MultiPolygon"]:
        for part in coords:
            for component in part:
                if isinstance(component[0], float | int):  # Handle MultiLineString
                    lon, lat = component
                    lons.append(lon)
                    lats.append(lat)
                else:  # Handle MultiPolygon
                    for lon, lat in component:
                        lons.append(lon)
                        lats.append(lat)

    else:
        msg = f"Unsupported geometry type: {geom_type}"
        raise ValueError(msg)

    return [min(lons), min(lats), max(lons), max(lats)]


def bbox_to_geojson(
    bbox: list[float] | tuple[float, float, float, float],
) -> dict[str, str | list[list[list[float]]]]:
    """
    Convert a bounding box into a Polygon geometry in dictionary format.

    Args:
        bbox (List[float]): A bounding box in the format [min_lat, min_lon, max_lat, max_lon].

    Returns:
        Dict[str, Union[str, List[List[List[float]]]]]: Polygon geometry represented as a dictionary.

    Note:
        The function doesn't use `shapely.geometry.shape` directly but constructs the polygon
        based on the provided bounding box.
    """

    return {
        "type": "Polygon",
        "coordinates": [
            [
                [bbox[2], bbox[1]],
                [bbox[2], bbox[3]],
                [bbox[0], bbox[3]],
                [bbox[0], bbox[1]],
                [bbox[2], bbox[1]],
            ]
        ],
    }



from .cloud import write_block, write_table
from .engine import STACQueryEngine
from .partitioner import EqualSizePartitioner
from .utils import name_data

__all__ = [
    "write_block",
    "write_table",
    "name_data",
    "STACQueryEngine",
    "EqualSizePartitioner",
]



import pathlib

import pystac
from pystac import Item
from pystac.layout import BestPracticesLayoutStrategy
from pystac.utils import JoinType, join_path_or_url, safe_urlparse


class ListItemLayout(BestPracticesLayoutStrategy):
    """
    Custom layout for CoCliCo STAC collections.

    This layout strategy modifies the default item path format. Instead of creating an item
    with the path format `/variable-mapbox-dim-value-dim-value/variable-mapbox-dim-value-dim-value.json`,
    it uses the format `variable-mapbox/variable-mapbox-dim-value-dim-value.json`.

    Attributes:
        item: The STAC item for which the href needs to be generated.
        parent_dir: The parent directory where the STAC item will reside.

    Returns:
        str: The href path for the given STAC item based on the custom layout.
    """

    def get_item_href(self, item: pystac.Item, parent_dir: str) -> str:
        """
        Generate the href path for a given STAC item based on the custom layout.

        Args:
            item (STACItem): The STAC item for which the href needs to be generated.
            parent_dir (str): The parent directory where the STAC item will reside.

        Returns:
            str: The href path for the given STAC item based on the custom layout.
        """
        parsed_parent_dir = safe_urlparse(parent_dir)
        join_type = JoinType.from_parsed_uri(parsed_parent_dir)
        items_dir = "items"
        custom_id = pathlib.Path(item.id).with_suffix(".json")
        return join_path_or_url(join_type, parent_dir, items_dir, str(custom_id))


class DynamicPrefixLayout(BestPracticesLayoutStrategy):
    """
    Custom layout for STAC collections based on dynamic prefixes from item IDs.

    This layout strategy generates item paths based on the first two segments of the
    item's ID. The directory is determined dynamically based on the item ID's structure,
    leading to a layout that looks like: `segment1-segment2/segment1-segment2-segment3-segment4.json`
    compared to a constant "items" directory.

    Attributes:
        item (pystac.Item): The STAC item for which the href needs to be generated.
        parent_dir (str): The parent directory where the STAC item will reside.

    Returns:
        str: The href path for the given STAC item based on the custom layout.
    """

    def get_item_href(self, item: pystac.Item, parent_dir: str) -> str:
        parsed_parent_dir = safe_urlparse(parent_dir)
        join_type = JoinType.from_parsed_uri(parsed_parent_dir)

        # Use the first two segments from the item's ID for the directory name
        custom_id = "-".join(item.id.split("-")[0:2])
        item_root = join_path_or_url(join_type, parent_dir, custom_id)
        return join_path_or_url(join_type, item_root, f"{item.id}.json")


class ParquetLayout(BestPracticesLayoutStrategy):
    """
    Custom layout for CoGs within CoCliCo STAC collections.

    Modifies the item path to:
        items/variable-mapbox-dim-value-dim-value.json
    Instead of the default:
        /variable-mapbox-dim-value-dim-value/variable-mapbox-dim-value-dim-value.json
    """

    def get_item_href(self, item: Item, parent_dir: str) -> str:
        """
        Determines the item href based on the custom layout for STAC items.

        Args:
            item (Item): The STAC item.
            parent_dir (str): The parent directory path.

        Returns:
            str: The constructed item href.
        """
        # Parse the parent directory URL or path
        parsed_parent_dir = safe_urlparse(parent_dir)
        join_type = JoinType.from_parsed_uri(parsed_parent_dir)

        # Use the `stac_item_id` directly from the item's ID, which is expected to be in the correct format
        stac_item_id = item.id

        # Define the directory where items will be stored relative to the parent directory
        items_dir = "items"

        # Construct the filename for the item's JSON representation
        filename = f"{stac_item_id}.json"

        # Join the parent directory, items directory, and filename to form the full item href
        return join_path_or_url(join_type, parent_dir, items_dir, filename)



import dataclasses
import datetime
import logging
import os
import pathlib
import re
from typing import Any

import fsspec
import pystac
import rasterio
import shapely
import stac_geoparquet
import xarray as xr

# uncomment these lines if you do not have coclicodata in development mode installed
# dev_dir = pathlib.Path.home() / "dev"  # set the path to the location where you would like to clone the package
# dev_dir.mkdir(parents=True, exist_ok=True)
# # Clone the repository
# os.system(f"git clone https://github.com/openearth/coclicodata.git {dev_dir / 'coclicodata'}")
# # Install the package in development mode
# os.system(f"pip install -e {dev_dir / 'coclicodata'}")
from coclicodata.coclico_stac.io import CoCliCoStacIO
from coclicodata.coclico_stac.layouts import CoCliCoCOGLayout
from dotenv import load_dotenv
from pystac.extensions import raster
from stactools.core.utils import antimeridian

# Load the environment variables from the .env file
load_dotenv(override=True)

logging.getLogger("azure").setLevel(logging.WARNING)

# Get the SAS token and storage account name from environment variables
sas_token = os.getenv("AZURE_STORAGE_SAS_TOKEN")
storage_account_name = os.getenv("AZURE_STORAGE_ACCOUNT_NAME")
storage_options = {"account_name": storage_account_name, "credential": sas_token}

# CoCliCo STAC
STAC_DIR = pathlib.Path.home() / "dev" / "coclicodata" / "current"

COLLECTION_ID = "deltares-delta-dtm"
COLLECTION_TITLE = "DeltaDTM: A global coastal digital terrain model"

ASSET_TITLE = "DeltaDTM"
ASSET_DESCRIPTION = "Deltares DeltaDTM with terrain elevation data for this region."
ASSET_EXTRA_FIELDS = {
    "xarray:storage_options": {"account_name": "coclico"},
}

DATE = datetime.datetime(2023, 10, 30, tzinfo=datetime.UTC)

NODATA_VALUE = -9999
RESOLUTION = 30

PARQUET_MEDIA_TYPE = "application/vnd.apache.parquet"

CONTAINER_NAME = "deltares-delta-dtm"
PREFIX = "v1.0.1"
CONTAINER_URI = f"az://{CONTAINER_NAME}/{PREFIX}"
GEOPARQUET_STAC_ITEMS_HREF = f"az://items/{COLLECTION_ID}.parquet"

EXAMPLE_HREF = "az://deltares-delta-dtm/v1.0.1/europe/DeltaDTM_v1_0_N03W052.tif"


@dataclasses.dataclass
class PathParser:
    """
    Parses a cloud storage path into its component parts, specifically designed for Azure Blob Storage and COG data.
    This class assumes paths are formatted like "az://<container>/<version>/<region>/<filename>.tif"
    """

    path: str
    container: str | None = None
    prefix: str | None = None
    name: str | None = None
    stac_item_id: str | None = None

    def __post_init__(self) -> None:
        stripped_path = re.sub(r"^\w+://", "", self.path)
        split_path = stripped_path.rstrip("/").split("/")

        self.container = split_path[0]
        self.name = split_path[-1]
        self.prefix = "/".join(split_path[1:-1])
        self.stac_item_id = self.name.rsplit(".", 1)[0]


def create_collection(
    description: str | None = None, extra_fields: dict[str, Any] | None = None
) -> pystac.Collection:
    providers = [
        pystac.Provider(
            name="Deltares",
            roles=[
                pystac.provider.ProviderRole.PRODUCER,
                pystac.provider.ProviderRole.PROCESSOR,
                pystac.provider.ProviderRole.HOST,
                pystac.provider.ProviderRole.LICENSOR,
            ],
            url="https://deltares.nl",
        ),
    ]

    extent = pystac.Extent(
        pystac.SpatialExtent([[-180.0, 90.0, 180.0, -90.0]]),
        pystac.TemporalExtent([[DATE, None]]),
    )

    links = [
        pystac.Link(
            pystac.RelType.LICENSE,
            target="https://creativecommons.org/licenses/by/4.0/",
            media_type="text/html",
            title="CC License",
        )
    ]

    keywords = [
        "Elevation",
        "...",
    ]

    description = """
    A global coastal digital terrain model, based on CopernicusDEM, ESA WorldCover, ICESat-2 and
    GEDI data. For more information, see Pronk et al. (2023) DeltaDTM: A global coastal digital terrain model.
    """

    collection = pystac.Collection(
        id=COLLECTION_ID,
        title=COLLECTION_TITLE,
        description=description,
        license="CC-BY-4.0",
        providers=providers,
        extent=extent,
        catalog_type=pystac.CatalogType.RELATIVE_PUBLISHED,
    )

    collection.add_asset(
        "thumbnail",
        pystac.Asset(
            "https://coclico.blob.core.windows.net/assets/thumbnails/deltares-delta-dtm-thumbnail.jpeg",
            title="Thumbnail",
            media_type=pystac.MediaType.JPEG,
        ),
    )
    collection.links = links
    collection.keywords = keywords

    pystac.extensions.item_assets.ItemAssetsExtension.add_to(collection)

    collection.extra_fields["item_assets"] = {
        "data": {
            "title": ASSET_TITLE,
            "description": ASSET_DESCRIPTION,
            "roles": ["data"],
            "type": pystac.MediaType.COG,
            **ASSET_EXTRA_FIELDS,
        }
    }

    if extra_fields:
        collection.extra_fields.update(extra_fields)

    pystac.extensions.scientific.ScientificExtension.add_to(collection)
    collection.extra_fields["sci:citation"] = (
        """Pronk, Maarten. 2023. “DeltaDTM: A Global Coastal Digital Terrain Model.” 4TU.ResearchData. https://doi.org/10.4121/21997565.V1."""
    )
    collection.extra_fields["sci:doi"] = "10.4121/21997565"
    collection.extra_fields["sci:publications"] = [
        {
            "doi": "10.1038/s41597-024-03091-9",
            "citation": """Pronk, Maarten, Aljosja Hooijer, Dirk Eilander, Arjen Haag, Tjalling de Jong, Michalis Vousdoukas, Ronald Vernimmen, Hugo Ledoux, and Marieke Eleveld. 2024. “DeltaDTM: A Global Coastal Digital Terrain Model.” Scientific Data 11 (1): 273. https://doi.org/10.1038/s41597-024-03091-9.""",
        }
    ]

    pystac.extensions.version.VersionExtension.add_to(collection)
    collection.extra_fields["version"] = "1.0"

    return collection


def create_item(block, item_id, antimeridian_strategy=antimeridian.Strategy.SPLIT):
    bbox = rasterio.warp.transform_bounds(block.rio.crs, 4326, *block.rio.bounds())
    geometry = shapely.geometry.mapping(shapely.make_valid(shapely.geometry.box(*bbox)))
    bbox = shapely.make_valid(shapely.box(*bbox)).bounds

    item = pystac.Item(
        id=item_id,
        geometry=geometry,
        bbox=bbox,
        datetime=DATE,
        properties={},
    )

    antimeridian.fix_item(item, antimeridian_strategy)

    item.common_metadata.created = datetime.datetime.utcnow()

    ext = pystac.extensions.projection.ProjectionExtension.ext(
        item, add_if_missing=True
    )
    ext.bbox = block.rio.bounds()  # these are provided in the crs of the data
    ext.shape = tuple(v for k, v in block.sizes.items() if k in ["y", "x"])
    ext.epsg = block.rio.crs.to_epsg()
    ext.geometry = shapely.geometry.mapping(shapely.geometry.box(*ext.bbox))
    ext.transform = list(block.rio.transform())[:6]
    ext.add_to(item)

    return item


def create_asset(
    item, asset_title, asset_href, nodata, resolution, data_type, nbytes=None
):
    asset = pystac.Asset(
        href=asset_href,
        media_type=pystac.MediaType.COG,
        title=asset_title,
        roles=["data"],
    )

    item.add_asset("data", asset)

    pystac.extensions.file.FileExtension.ext(asset, add_if_missing=True)

    if nbytes:
        asset.extra_fields["file:size"] = nbytes

    raster.RasterExtension.ext(asset, add_if_missing=True).bands = [
        raster.RasterBand.create(
            nodata=nodata,
            spatial_resolution=resolution,
            data_type=data_type,  # e.g., raster.DataType.INT8
        )
    ]

    return item


if __name__ == "__main__":
    fs, token, [root] = fsspec.get_fs_token_paths(
        CONTAINER_URI, storage_options=storage_options
    )

    fps = fs.glob(f"{root}/**/*.tif")

    stac_io = CoCliCoStacIO()
    layout = CoCliCoCOGLayout()

    collection = create_collection()

    # NOTE: adjust logging level to avoid this warning:
    # INFO:rasterio._filepath:Object not found in virtual filesystem:
    # filename=b'ed4e8723-89d8-4075-b946-1ae1a21cca03/ed4e8723-89d8-4075-b946-1ae1a21cca03.aux'
    logging.getLogger("rasterio").setLevel(logging.WARNING)

    for fp in fps:
        href = "az://" + fp
        pp = PathParser(href)

        with fs.open(fp, "rb") as f:
            block = xr.open_dataset(f, engine="rasterio")["band_data"].squeeze()

        item = create_item(block, pp.stac_item_id)
        item = create_asset(
            item,
            ASSET_TITLE,
            pp.path,
            nodata=NODATA_VALUE,
            resolution=RESOLUTION,
            data_type=raster.DataType.FLOAT32,
        )
        collection.add_item(item)

    collection.update_extent_from_items()

    items = list(collection.get_all_items())
    items_as_json = [i.to_dict() for i in items]
    item_extents = stac_geoparquet.to_geodataframe(items_as_json)

    with fsspec.open(GEOPARQUET_STAC_ITEMS_HREF, mode="wb", **storage_options) as f:
        item_extents.to_parquet(f)

    collection.add_asset(
        "geoparquet-stac-items",
        pystac.Asset(
            GEOPARQUET_STAC_ITEMS_HREF,
            title="GeoParquet STAC items",
            description="Snapshot of the collection's STAC items exported to GeoParquet format.",
            media_type=PARQUET_MEDIA_TYPE,
            roles=["data"],
        ),
    )

    catalog = pystac.Catalog.from_file(str(STAC_DIR / "catalog.json"))

    if catalog.get_child(collection.id):
        catalog.remove_child(collection.id)
        print(f"Removed child: {collection.id}.")

    catalog.add_child(collection)

    collection.normalize_hrefs(str(STAC_DIR / collection.id), strategy=layout)

    collection.validate_all()

    catalog.save(
        catalog_type=pystac.CatalogType.SELF_CONTAINED,
        dest_href=str(STAC_DIR),
        stac_io=stac_io,
    )

    catalog.validate_all()



import copy
import json
import logging
import pathlib
import uuid
import warnings
from datetime import datetime
from posixpath import join as urljoin
from urllib.parse import urlsplit

import fsspec
import geopandas as gpd
import pandas as pd
import xarray
import xarray as xr
from pyproj import Transformer
from shapely.geometry import box
from shapely.ops import transform

logger = logging.getLogger(__name__)


def is_local_file_path(path: str | pathlib.Path) -> bool:
    """
    Determine if a given path is a local filesystem path using urlsplit.

    Args:
        path (str): The path to check.

    Returns:
        bool: True if it's a local file path, False otherwise.
    """
    parsed = urlsplit(str(path))

    return parsed.scheme in ["", "file"]


def name_block(
    block: xr.DataArray,
    storage_prefix: str = "",
    name_prefix: str | None = None,
    include_band: str | None = None,
    time_dim: str | None = "time",
    x_dim: str | None = "x",
    y_dim: str | None = "y",
) -> str:
    """Create a name for a block based on its coordinates and other options.

    Args:
        block (xr.DataArray): The DataArray block whose name is to be generated.
        storage_prefix (str | None, optional): Storage prefix to prepend to blob name. Defaults to None.
        name_prefix (Optional[str], optional): String prefix to prepend to the name. Defaults to None.
        include_band (Optional[str], optional): Name of the band to include in the name. Defaults to None.
        time_dim (Optional[str], optional): Name of the time dimension. Defaults to "time".
        x_dim (str | None, optional): Name of the x dimension. Defaults to "x".
        y_dim (str | None, optional): Name of the y dimension. Defaults to "y".

    Returns:
        str: A string name constructed from the block's coordinates.

    Raises:
        ValueError: If no valid components are available to form a name.
    """

    storage_prefix = str(storage_prefix) if storage_prefix else ""

    components = []

    epsg = block.rio.crs.to_epsg()

    if name_prefix:
        components.append(name_prefix)

    if (x_dim or y_dim) and (epsg != 4326 and epsg != 3857):
        components.append(f"epsg={block.rio.crs.to_epsg()}")

    if include_band:
        components.append(include_band)

    if time_dim and time_dim in block.coords:
        time = pd.Timestamp(block.coords[time_dim].item()).isoformat()
        components.append(f"{time_dim}={time}")

    if x_dim and x_dim in block.coords:
        if epsg == 4326:
            x_val = round(block.coords[x_dim].min().item(), 2)
        else:
            x_val = int(round(block.coords[x_dim].min().item(), 2))
        components.append(f"{x_dim}={x_val}")

    if y_dim and y_dim in block.coords:
        if epsg == 4326:
            y_val = round(block.coords[y_dim].min().item(), 2)
        else:
            y_val = int(round(block.coords[y_dim].min().item(), 2))
        components.append(f"{y_dim}={y_val}")

    if not components:
        msg = (
            "No valid components to form a name. Check the input DataArray and provided"
            " arguments."
        )
        raise ValueError(msg)

    name = f"{'_'.join(components)}.tif"

    if not is_local_file_path(storage_prefix):  # cloud storage
        return str(urljoin(storage_prefix, name))
    else:  # local storage
        return str(pathlib.Path(storage_prefix) / name)


def name_table(
    gdf: gpd.GeoDataFrame,
    storage_prefix: str = "",
    name_prefix: str | None = None,
    x_dim: str = "x",
    y_dim: str = "y",
) -> str:
    """Create a name for a Parquet file based on the total_bounds of a GeoDataFrame.

    Args:
        gdf (gpd.GeoDataFrame): The GeoDataFrame whose file name is to be generated.
        storage_prefix (str, optional): Storage prefix to prepend to file name. Defaults to "".
        name_prefix (str | None, optional): String prefix to prepend to the name. Defaults to None.
        x_dim (str, optional): Name of the x dimension. Defaults to "x".
        y_dim (str, optional): Name of the y dimension. Defaults to "y".

    Returns:
        str: A string name constructed from the GeoDataFrame's total bounds.

    Raises:
        ValueError: If no valid components are available to form a name.
    """

    storage_prefix = str(storage_prefix) if storage_prefix else ""

    components = []

    if name_prefix:
        components.append(name_prefix)

    epsg = gdf.crs.to_epsg()
    if epsg and epsg not in [4326, 3857]:
        components.append(f"epsg={epsg}")
        warnings.warn(
            "It is recommended to save GeoSpatial tabular data in EPSG 4326 or 3857 to"
            " ease global analyses.",
            stacklevel=2,
        )

    # Using total bounds for naming
    try:
        total_bounds = gdf.total_bounds
        x_min, y_min, _, _ = map(int, total_bounds)
        components.append(f"{x_dim}={x_min}_{y_dim}={y_min}")
    except ValueError:
        components.append("")

    if not components:
        msg = (
            "No valid components to form a name. Check the input GeoDataFrame and"
            " provided arguments."
        )
        raise ValueError(msg)

    name = f"{'_'.join(components)}.parquet"

    if not is_local_file_path(storage_prefix):
        return str(urljoin(storage_prefix, name))
    else:
        return str(pathlib.Path(storage_prefix) / name)


def name_data(
    data: pd.DataFrame | gpd.GeoDataFrame | xarray.Dataset | xarray.DataArray,
    prefix: str | None = None,
    filename_prefix: str | None = None,
    include_bounds: bool = True,
    include_random_hex: bool = True,
) -> str:
    """
    Generates a unique filename for a dataset with optional geographic bounds and random hex string.
    If neither bounds nor hex string are included, raises a ValueError.

    Args:
        data: A DataFrame, GeoDataFrame, or xarray dataset/data array representing spatial or non-spatial data.
        prefix: An optional prefix for the filename, typically used for higher-level grouping.
        filename_prefix: An additional prefix for the filename, used for further categorization.
        include_bounds: Flag to include geographic bounds in the filename for geospatial datasets.
        include_random_hex: Flag to include a random hex string in the filename.
        include_random_hex: The zoom level of the QuadKey prefix to include.

    Returns:
        A string representing the uniquely generated filename based on the provided options.

    Raises:
        ValueError: If both include_bounds and include_random_hex are False, or if the data type is unsupported.
    """
    if not include_bounds and not include_random_hex:
        msg = "At least one of include_bounds, include_random_hex or quadkey_prefix must be set."
        raise ValueError(msg)

    parts = []

    # Generate bounds part
    bounds_part = ""
    if isinstance(data, pd.DataFrame | gpd.GeoDataFrame):
        suffix = ".parquet"

    if include_bounds and isinstance(
        data, gpd.GeoDataFrame | xarray.Dataset | xarray.DataArray
    ):
        if isinstance(data, gpd.GeoDataFrame):
            bounds = data.total_bounds
            crs = data.crs.to_string()

        else:
            suffix = ".tif"
            bounds = data.rio.bounds()
            crs = data.rio.crs.to_string()

        def name_bounds(bounds, crs):
            bounds_geometry = box(*bounds)
            if crs != "EPSG:4326":
                transformer = Transformer.from_crs(crs, "EPSG:4326", always_xy=True)
                bounds_geometry = transform(transformer.transform, bounds_geometry)
            minx, miny, _, _ = bounds_geometry.bounds
            lon_prefix = "e" if minx >= 0 else "w"
            lat_prefix = "n" if miny >= 0 else "s"
            formatted_minx = f"{abs(minx):03.0f}"
            formatted_miny = f"{abs(miny):02.0f}"
            return f"{lat_prefix}{formatted_miny}{lon_prefix}{formatted_minx}"

        bounds_part = name_bounds(bounds, crs)
        if bounds_part:
            parts.append(bounds_part)

    # Generate random hex part
    random_hex = ""
    if include_random_hex:
        random_hex = uuid.uuid4().hex[:12]
        parts.append(f"part-{random_hex}")

    if not parts:
        msg = "No valid parts were generated for the filename."
        raise ValueError(msg)

    name_part = "_".join(parts) + suffix

    # Constructing the final filename
    components = [comp for comp in [filename_prefix, name_part] if comp]
    filename = "_".join(components)

    if prefix:
        return f"{prefix}/{filename}"

    return filename


def read_items_extent(collection, columns=None, storage_options=None):
    """
    Reads the extent of items from a STAC collection and returns a GeoDataFrame with specified columns.

    Args:
        collection: A STAC collection object that contains assets.
        columns: List of columns to return. Default is ["geometry", "assets", "href"].
        storage_options: Storage options to pass to fsspec. Default is None.

    Returns:
        GeoDataFrame containing the specified columns.
    """
    if storage_options is None:
        storage_options = {}

    # Set default columns
    if columns is None:
        columns = ["geometry", "assets", "href"]

    columns_ = copy.deepcopy(columns)

    # Ensure 'assets' is always in the columns
    if "assets" not in columns:
        columns.append("assets")
        logger.debug("'assets' column added to the list of columns")

    # Open the parquet file and read the specified columns
    href = collection.assets["geoparquet-stac-items"].href
    with fsspec.open(href, mode="rb", **storage_options) as f:
        extents = gpd.read_parquet(f, columns=[c for c in columns if c != "href"])

    # If 'href' is requested, extract it from the 'assets' column
    if "href" in columns:
        extents["href"] = extents["assets"].apply(lambda x: x["data"]["href"])
        logger.debug("'href' column extracted from 'assets'")

    # Drop 'assets' if it was not originally requested
    if "assets" not in columns_:
        extents = extents.drop(columns=["assets"])
        logger.debug("'assets' column dropped from the GeoDataFrame")

    return extents


def write_log_entry(
    container: str,
    name: str,
    status: str,
    storage_options: dict[str, str] | None = None,
    prefix: str | None = None,
    time: str | datetime | None = None,
) -> None:
    """
    Adds a new entry to a JSON log file with a given name (can be URI), time, and status.
    The log file is stored in the specified container and optionally prefixed with the given prefix.

    Args:
        container (str): The base path to the log container.
        name (str): The name to log.
        status (str): The status of the processing (e.g., 'success', 'failed').
        storage_options (Optional[Dict[str, str]]): Authentication and configuration options for Azure storage.
        prefix (Optional[str]): The prefix for the log file path.
        time (Optional[Union[str, datetime]]): The time of logging. If None, the current time is used.
    """
    if storage_options is None:
        storage_options = {}
    # Generate a UUID for the log file name
    log_id = uuid.uuid4().hex[:16]
    log_filename = f"{log_id}.json"

    # Construct the full path to the log file
    log_path = (
        f"{container}/{prefix}/{log_filename}"
        if prefix
        else f"{container}/{log_filename}"
    )

    # Ensure the time is in ISO format
    if time is None:
        time = datetime.now().isoformat()
    else:
        try:
            if isinstance(time, str):
                time = datetime.fromisoformat(time).isoformat()
            elif isinstance(time, datetime):
                time = time.isoformat()
            else:
                raise ValueError
        except ValueError:
            time = datetime.now().isoformat()

    # Create the log entry
    log_entry = {"name": name, "time": time, "status": status}

    # Write the log entry to the specified path in JSON format
    with fsspec.open(log_path, "w", **storage_options) as f:
        json.dump(log_entry, f)


def read_log_entries(
    base_uri: str, storage_options: dict[str, str | None], prefix: str | None = None
) -> pd.DataFrame:
    """
    Lists all JSON log files under a specified prefix, reads them into a DataFrame,
    and sorts them by ascending time.

    Args:
        base_uri (str): The base URI to search for JSON log files.
        storage_options (Dict[str, str]): Authentication and configuration options for Azure storage.
        prefix (Optional[str]): A prefix to filter the log files.

    Returns:
        pd.DataFrame: A DataFrame of log entries sorted by ascending date.
    """
    protocol = base_uri.split("://")[0]
    fs = fsspec.filesystem(protocol, **storage_options)

    # Use glob to list all JSON files under the base URI with the optional prefix
    search_pattern = f"{base_uri}/{prefix}/*.json" if prefix else f"{base_uri}/*.json"
    json_files = fs.glob(search_pattern)

    # Read JSON files into a list of dictionaries
    logs: list[dict] = []
    for f in json_files:
        with fs.open(f, "r", **storage_options) as f2:
            log_entry = json.load(f2)
            logs.append(log_entry)

    # If no log entries are found, return an empty DataFrame
    if not logs:
        return pd.DataFrame(columns=["name", "time", "status"])

    # Create a DataFrame from the list of dictionaries
    df = pd.DataFrame(logs)

    # Ensure the time column is in datetime format, raise an error if parsing fails
    try:
        df["time"] = pd.to_datetime(df["time"], errors="raise")
    except Exception as e:
        msg = f"Error parsing datetime: {e}"
        raise ValueError(msg)  # noqa: B904

    # Sort DataFrame by time in ascending order
    df = df.sort_values(by="time", ascending=True)

    return df


def rm_from_storage(
    pattern: str,
    storage_options: dict[str, str] | None = None,
    confirm: bool = True,
    verbose: bool = True,
) -> None:
    """
    Deletes all blobs/files in the specified storage location that match the given prefix.

    Args:
        pattern (str): The pattern or path pattern (including wildcards) for the blobs/files to delete.
        storage_options (Dict[str, str], optional): A dictionary containing storage connection details.
        confirm (bool): Whether to prompt for confirmation before deletion.
        verbose (bool): Whether to display detailed log messages.

    Returns:
        None
    """
    if storage_options is None:
        storage_options = {}

    # Create a local logger
    logger = logging.getLogger(__name__)
    if verbose:
        handler = logging.StreamHandler()
        handler.setLevel(logging.INFO)
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)

    if storage_options is None:
        storage_options = {}

    # Get filesystem, token, and resolved paths
    fs, _, paths = fsspec.get_fs_token_paths(pattern, storage_options=storage_options)

    if paths:
        if verbose:
            logger.info(
                f"Warning: You are about to delete the following {len(paths)} blobs/files matching '{pattern}'."
            )
            for path in paths:
                logger.info(path)

        if confirm:
            confirmation = input(
                f"Type 'yes' to confirm deletion of {len(paths)} blobs/files matching '{pattern}': "
            )
        else:
            confirmation = "yes"

        if confirmation.lower() == "yes":
            for path in paths:
                try:
                    if verbose:
                        logger.info(f"Deleting blob/file: {path}")
                    fs.rm(path)
                    if verbose:
                        logger.info(f"Blob/file {path} deleted successfully.")
                except Exception as e:
                    if verbose:
                        logger.error(f"Failed to delete blob/file: {e}")
            if verbose:
                logger.info("All specified blobs/files have been deleted.")
        else:
            if verbose:
                logger.info("Blob/file deletion cancelled.")
    else:
        if verbose:
            logger.info(f"No blobs/files found matching '{pattern}'.")

    # Remove the handler after use
    if verbose:
        logger.removeHandler(handler)
        handler.close()



from typing import Any

import geopandas as gpd
import shapely
from shapely import LineString, Polygon

from coastpy.geo.ops import generate_offset_line
from coastpy.geo.transform import (
    bbox_to_geojson,
)


def bbox_from_coords(coords: list[float | list[Any]]) -> list[float]:
    """
    Extract the bounding box from a list of coordinates.

    This function recursively goes through a list of coordinates, which can be
    nested to various levels, and returns the bounding box for the given coordinates.

    Args:
        coords (List[Union[float, List[Any]]]): A list of coordinates which can be
            nested lists representing complex geometries.

    Returns:
        List[float]: Bounding box in the format [min_lat, min_lon, max_lat, max_lon].

    Raises:
        AssertionError: If there's a type mismatch in the coordinates.
    """

    lats, lons = [], []

    def _extract(coords: list[float | int | list[Any]]) -> None:
        for x in coords:
            # This handles points
            if isinstance(x, float):
                assert isinstance(
                    coords[0], float
                ), f"Type mismatch: {coords[0]} is not a float"
                assert isinstance(
                    coords[1], float
                ), f"Type mismatch: {coords[1]} is not a float"
                lats.append(coords[0])
                lons.append(coords[1])
                return
            # This handles nested lists of coordinates
            if isinstance(x, list) and isinstance(x[0], list):
                _extract(x)
            else:
                lat, lon = x
                lats.append(lat)
                lons.append(lon)

    _extract(coords)
    lats.sort()
    lons.sort()

    return [lats[0], lons[0], lats[-1], lons[-1]]


def geo_bbox(
    min_lon: float,
    min_lat: float,
    max_lon: float,
    max_lat: float,
    src_crs: str = "EPSG:4326",
    dst_crs: str = "EPSG:4326",
) -> gpd.GeoDataFrame:
    """GeoDataFrame with a bounding box that can be used for various gis operations.
    Args:
        min_lon (float): most eastern longitude
        min_lat (float): most northern latitude
        max_lon (float): most wester longitude
        max_lat (float): most soutern latitude
        src_crs (str, optional): Valid EPSG string or number (int). Defaults to "EPSG:4326".
        dst_crs (str, optional): Valid EPSG string or number (int). Defaults to "EPSG:4326".

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with bounding box as geometry.
    """

    bbox = [min_lon, min_lat, max_lon, max_lat]
    bbox = bbox_to_geojson(bbox)
    bbox = shapely.geometry.shape(bbox)
    return gpd.GeoDataFrame(geometry=[bbox], crs=src_crs).to_crs(dst_crs)


def create_offset_rectangle(line: LineString, distance: float) -> Polygon:
    """
    Construct a rectangle polygon using the original line and an offset distance.

    Args:
        line (LineString): The original line around which the polygon is constructed.
        distance (float): The offset distance used to create the sides of the polygon.

    Returns:
        Polygon: The constructed rectangle-shaped polygon.
    """

    # Create the offset lines
    left_offset_line = generate_offset_line(line, distance)
    right_offset_line = generate_offset_line(line, -distance)

    # Retrieve end points
    left_start, left_end = left_offset_line.coords[:]
    right_start, right_end = right_offset_line.coords[:]

    # Construct the polygon using the end points
    polygon = Polygon([left_start, left_end, right_end, right_start])

    return polygon



import math
from itertools import product, repeat

import geopandas as gpd
import mercantile
import numpy as np
import shapely.geometry as sgeom
from mercantile import Tile
from shapely.geometry import Polygon, box


def tile_to_quadkey(tile: Tile) -> str:
    """
    Convert a Tile to its corresponding quadkey.

    Args:
        tile (Tile): A mercantile Tile object.

    Returns:
        str: The quadkey corresponding to the tile.
    """
    quadkey = []
    for i in range(tile.z, 0, -1):
        digit = 0
        mask = 1 << (i - 1)
        if (tile.x & mask) != 0:
            digit += 1
        if (tile.y & mask) != 0:
            digit += 2
        quadkey.append(str(digit))
    return "".join(quadkey)


def make_mercantiles(zoom_level: int) -> gpd.GeoDataFrame:
    """
    Generate a GeoDataFrame of tiles for a given zoom level with their corresponding quadkeys.

    Args:
        zoom_level (int): The zoom level for which to generate tiles.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing tile geometries and their quadkeys.
    """
    tiles_data: list[dict[str, str]] = []
    num_tiles = 2**zoom_level

    for x in range(num_tiles):
        for y in range(num_tiles):
            tile = mercantile.Tile(x=x, y=y, z=zoom_level)
            bounds = mercantile.bounds(tile)
            tile_polygon = box(bounds.west, bounds.south, bounds.east, bounds.north)
            quadkey = mercantile.quadkey(tile)
            tiles_data.append({"geometry": tile_polygon, "quadkey": quadkey})

    df = gpd.GeoDataFrame(tiles_data, columns=["geometry", "quadkey"], crs=4326)
    return df


def deg2num(lon_deg: float, lat_deg: float, zoom: int) -> tuple:
    """Derive tile number for point in WGS 84 given certain zoom level.

    Args:
        lon_deg (float): Longitude as float in WGS 84.
        lat_deg (float):Latitude as float in WGS 84.
        zoom (int): Zoom following OSM slippy tile names conventions.

    Returns:
        _type_: tile number (xtile, ytile) following OSM slippy tile names conventions.
    """

    lat_rad = math.radians(lat_deg)
    n = 2.0**zoom
    xtile = int((lon_deg + 180.0) / 360.0 * n)
    ytile = int((1.0 - math.asinh(math.tan(lat_rad)) / math.pi) / 2.0 * n)
    return (xtile, ytile)


def num2deg(xtile: int, ytile: int, zoom: int) -> tuple:
    """Tile number (xtile, ytile) to WGS 84 point (longitude, latitude) of northwest tile corner.

    Args:
        xtile (int): xtile number as int.
        ytile (int): ytile number as int int.
        zoom (int): zoom level following OSM slippy map names convention.

    Returns:
        tuple: point (longitude, latitude) in WGS 84.
    """
    n = 2.0**zoom
    lon_deg = xtile / n * 360.0 - 180.0
    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
    lat_deg = math.degrees(lat_rad)
    return (lon_deg, lat_deg)


def unique_items(tile_numbers: list) -> list:
    """Filter unique tuples from list of tuples.

    Args:
        tile_numbers (list): List of tuples.

    Returns:
        list: List of tuples.
    """
    seen = set()
    result = []
    for item in tile_numbers:
        if item not in seen:
            seen.add(item)
            result.append(item)
    return result


def coords2num(lons: list, lats: list, zoom: int) -> list:
    """List of longitude / latitude coordinates to OpenStreetMap tile names given certain zoom level.

    Args:
        lons (list): List of longitude as floats in WGS 84.
        lats (list): List of latitudes as floats in WGS 84.
        zoom (int): Zoom level following OSM slippy tile names.

    Returns:
        list: List of tuples with tile numbers (xtile, ytile).
    """
    tile_numbers = list(map(deg2num, lons, lats, repeat(zoom)))

    return unique_items(tile_numbers)


def zoom2num(zoom: int) -> list:
    """Generate tile numbers for zoom level.

    Args:
        zoom (int): zoom level in accordance with OpenStreetMap slippy tile names.

    Returns:
        list: list with tuples of tile numbers.
    """
    # tiles (zoom level) = 2**zoom * 2**zoom
    ntiles = 2**zoom
    tile_arange = np.arange(ntiles)
    return list(product(tile_arange, tile_arange))


def tile2box(xtile: int, ytile: int, zoom: int) -> dict:
    """Generate box for tile (xtile, ytile) at specified zoom level following OpenStreetMap
    slippy tile names conventions.

    Args:
        xtile int: xtile coordinate
        ytile int: ytile coordinate (int)
        zoom int: zoom level

    Returns:
        dict: with box name (zoom + xtile + ytile) as key and box (shapely.geometry.Polygon) as value.
    """
    nw = num2deg(xtile, ytile, zoom)
    ne = num2deg(xtile + 1, ytile, zoom)
    sw = num2deg(xtile, ytile + 1, zoom)
    se = num2deg(xtile + 1, ytile + 1, zoom)
    boxname = f"z{zoom}x{xtile}y{ytile}"
    return {boxname: sgeom.Polygon([nw, ne, sw, se]).envelope}


def slippy_map_tiles_to_quadkey(
    xtile: int, ytile: int, zoom: int
) -> dict[str, Polygon]:
    """Convert tile coordinates to a bounding box with an associated quadkey.

    This function computes the bounding box for a given tile, represented by its
    xtile, ytile, and zoom values. The resulting box is associated with a quadkey
    following the Bing Maps Tile System.

    Args:
        xtile (int): The x-coordinate of the tile.
        ytile (int): The y-coordinate of the tile.
        zoom (int): The zoom level.

    Returns:
        dict: A dictionary with the quadkey as the key and the bounding box (as a
        shapely Polygon) as the value.
    """

    def _quadkey(x: int, y: int, z: int) -> str:
        """Convert tile coordinates to a quadkey."""
        quadkey = ""
        for i in range(z, 0, -1):
            digit = 0
            mask = 1 << (i - 1)
            if (x & mask) != 0:
                digit += 1
            if (y & mask) != 0:
                digit += 2
            quadkey += str(digit)
        return quadkey

    nw = num2deg(xtile, ytile, zoom)
    ne = num2deg(xtile + 1, ytile, zoom)
    sw = num2deg(xtile, ytile + 1, zoom)
    se = num2deg(xtile + 1, ytile + 1, zoom)

    quadkey = _quadkey(xtile, ytile, zoom)
    return {quadkey: Polygon([nw, ne, sw, se]).envelope}


def make_slippy_map_tiles(zoom_level: int, crs: str) -> gpd.GeoDataFrame:
    """Generate boxes for a given zoom level.

    Args:
        zoom_level (int): Zoom level in accordance with OSM slippy tile conventions.
        crs (str): Coordinate reference system string.

    Returns:
        gpd.GeoDataFrame: Geodataframe containing the boxes.
    """
    tile_numbers = zoom2num(zoom_level)

    # Convert the tile coordinates to boxes by deriving the corners.
    boxes = [tile2box(*xytile, zoom=zoom_level) for xytile in tile_numbers]

    # Merge the list of dictionaries into one dict.
    boxes = {k: v for d in boxes for k, v in d.items()}

    # Load into GeoPandas and convert to desired CRS
    boxes = gpd.GeoDataFrame(
        boxes.items(), columns=["tilekey", "geometry"], crs="EPSG:4326"
    ).to_crs(crs)
    return boxes



from pathlib import Path
from typing import Literal

import dask_geopandas
import fsspec
import mercantile

from coastpy.geo.quadtiles_utils import add_geo_columns
from coastpy.geo.size import estimate_memory_usage_per_row
from coastpy.io.utils import name_data
from coastpy.utils.size import size_to_bytes


# NOTE: What about separating the partitioner and the equal size partitioner? - that
# way we can have a more generic partitioner tha does not have to compute the sizes of
# each row.
class EqualSizePartitioner:
    def __init__(
        self,
        df,
        out_dir,
        max_size,
        sort_by="quadkey",
        quadkey_zoom_level=12,
        geo_columns: list[Literal["bbox", "bounding_quadkey", "quadkey"]] | None = None,
        column_order=None,
        dtypes=None,
        storage_options=None,
        naming_function_kwargs=None,
    ):
        if geo_columns is None:
            geo_columns = ["bbox", "quadkey"]

        if storage_options is None:
            storage_options = {}

        if naming_function_kwargs is None:
            naming_function_kwargs = {}

        self.df = df
        self.out_dir = out_dir
        self.max_size_bytes = size_to_bytes(max_size)
        self.sort_by = sort_by
        self.quadkey_zoom_level = quadkey_zoom_level
        self.geo_columns = geo_columns
        self.storage_options = storage_options
        self.column_order = column_order
        self.dtypes = dtypes
        self.naming_function_kwargs = naming_function_kwargs

        # Set the naming function for the output files
        self.naming_function = name_data

    def preprocess(self):
        """
        Preprocess the dataframe by estimating row size and cumulative sum.
        Also, ensure necessary columns like 'quadkey', 'bbox', and 'bounding_quadkey' are present.
        """
        df = self.df.copy()  # Work on a copy to avoid modifying the original dataframe
        df["size"] = estimate_memory_usage_per_row(df)

        if not all(column in df.columns for column in self.geo_columns):
            df = add_geo_columns(
                df,
                geo_columns=self.geo_columns,
                quadkey_zoom_level=self.quadkey_zoom_level,
            )

        df = df.sort_values(self.sort_by).reset_index(drop=True)
        df["size_cumsum"] = df["size"].cumsum()

        return df

    def process(self):
        """
        Process each group of data based on 'quadkey'.
        """
        df = self.preprocess()
        while not df.empty and df["size_cumsum"].max() > self.max_size_bytes:
            split_index = df[df["size_cumsum"] > self.max_size_bytes].index.min() + 1
            partition_df = df.iloc[:split_index].reset_index(drop=True)
            df = df.iloc[split_index:].reset_index(drop=True)
            df["size_cumsum"] = df["size"].cumsum()

            if not partition_df.empty:
                self.write_data(partition_df)

        if not df.empty:
            self.write_data(df)

    def write_data(self, partition_df, column_order=None):
        """
        Write the processed dataframe partition to a parquet file using fsspec
        to support both local and cloud storage.
        """
        if not partition_df.empty:
            partition_df = partition_df[column_order] if column_order else partition_df
            # Generate the output path using the naming function
            outpath = self.naming_function(
                partition_df, prefix=str(self.out_dir), **self.naming_function_kwargs
            )

            # Initialize the filesystem object with storage options

            fs, _, [path] = fsspec.get_fs_token_paths(
                outpath, storage_options=self.storage_options
            )

            # Ensure the output directory exists (local filesystem specific)
            if fs.protocol == ("file", "local"):
                # For local filesystem, ensure the parent directory exists
                fs.makedirs(fs._parent(path), exist_ok=True)

            if self.dtypes:
                partition_df = partition_df.astype(self.dtypes)

            if self.column_order:
                partition_df = partition_df[self.column_order]

            # Use fsspec to write the DataFrame to parquet
            with fs.open(path, "wb") as f:
                partition_df.to_parquet(f, index=False)


class QuadKeyEqualSizePartitioner(EqualSizePartitioner):
    def __init__(
        self,
        df,
        out_dir,
        max_size,
        min_quadkey_zoom,
        sort_by,
        quadkey_zoom_level=12,
        geo_columns: list[Literal["bbox", "bounding_quadkey", "quadkey"]] | None = None,
        column_order=None,
        dtypes=None,
        storage_options=None,
        naming_function_kwargs=None,
    ):
        super().__init__(
            df,
            out_dir,
            max_size,
            sort_by,
            quadkey_zoom_level,
            geo_columns,
            column_order,
            dtypes,
            storage_options,
            naming_function_kwargs,
        )
        self.min_quadkey_zoom = min_quadkey_zoom
        self.quadkey_grouper = f"quadkey_z{min_quadkey_zoom}"
        self.column_order = column_order
        self.dtypes = dtypes

    def add_quadkey_group(self):
        """
        Adds a column to the DataFrame with quadkey based on the specified zoom level.
        """
        self.df[self.quadkey_grouper] = self.df.apply(
            lambda r: mercantile.quadkey(
                mercantile.tile(r.lon, r.lat, self.min_quadkey_zoom)
            ),
            axis=1,
        )

    def process(self):
        """
        Override the process method to handle quadkey-based partitioning.
        """
        self.add_quadkey_group()
        grouped = self.df.groupby(self.quadkey_grouper)

        for _, group in grouped:
            group2 = group.drop(columns=[self.quadkey_grouper], errors="ignore")
            partitioner = EqualSizePartitioner(
                group2,
                self.out_dir,
                self.max_size_bytes,
                self.sort_by,
                self.quadkey_zoom_level,
                self.geo_columns,
                self.column_order,
                self.dtypes,
            )
            partitioner.process()

    def write_data(self, partition_df):
        """
        Writes each partition to a separate parquet file.
        """
        super().write_data(partition_df)


class HivePartitionEqualSizePartitioner:
    def __init__(self, data_dir, out_dir, max_size, dtypes):
        self.data_dir = Path(data_dir)
        self.out_dir = Path(out_dir)
        self.max_size = max_size
        self.out_dir.mkdir(parents=True, exist_ok=True)
        self.dtypes = dtypes

    def read_hive_partitions(self):
        """
        Read data from Hive-partitioned directories and aggregate into a single DataFrame.
        """
        pattern = str(self.data_dir / "*=*/" / "*.parquet")
        df = dask_geopandas.read_parquet(pattern, engine="pyarrow")
        return df

    def process(self):
        """
        Process data from Hive-partitioned directories for equal-size partitioning.
        """
        df = self.read_hive_partitions().compute()
        partitioner = EqualSizePartitioner(df, self.out_dir, self.max_size, self.dtypes)
        partitioner.process()



from typing import Literal

import geopandas as gpd
import mercantile
from shapely import Point
from shapely.geometry import shape


def bbox_to_quadkey(
    min_lon: float, min_lat: float, max_lon: float, max_lat: float
) -> str | None:
    """
    Converts a bounding box to a single quadkey, if possible, for the highest zoom level where the bounding box
    fits into a single tile.

    Args:
    - min_lon (float): The minimum longitude of the bounding box.
    - min_lat (float): The minimum latitude of the bounding box.
    - max_lon (float): The maximum longitude of the bounding box.
    - max_lat (float): The maximum latitude of the bounding box.

    Returns:
    - Optional[str]: The quadkey string if the bounding box fits into a single tile at any zoom level from 12 down to 0,
      None otherwise.
    """
    for zoom in range(12, -1, -1):
        tiles = list(mercantile.tiles(min_lon, min_lat, max_lon, max_lat, zooms=zoom))
        if len(tiles) == 1:
            return mercantile.quadkey(tiles[0])
    return None


def geojson_to_quadkey(geojson_data: dict) -> str | None:
    """
    Converts GeoJSON geometry to a quadkey string, if the geometry fits into a single tile at any zoom level.

    Args:
    - geojson_data (Dict): The GeoJSON data containing the geometry.

    Returns:
    - Optional[str]: The quadkey string if the geometry fits into a single tile, None otherwise.
    """
    geom = shape(geojson_data["geometry"])
    min_lon, min_lat, max_lon, max_lat = geom.bounds
    return bbox_to_quadkey(min_lon, min_lat, max_lon, max_lat)


def quadkey_to_geojson(quadkey: str) -> dict:
    """
    Converts a quadkey string to its GeoJSON polygon representation.

    Args:
    - quadkey (str): The quadkey string representing a single tile.

    Returns:
    - Dict: The GeoJSON representation of the tile's bounding box as a polygon.
    """
    tile = mercantile.quadkey_to_tile(quadkey)
    bbox = mercantile.bounds(tile)
    return {
        "type": "Feature",
        "geometry": {
            "type": "Polygon",
            "coordinates": [
                [
                    [bbox.west, bbox.south],
                    [bbox.east, bbox.south],
                    [bbox.east, bbox.north],
                    [bbox.west, bbox.north],
                    [bbox.west, bbox.south],
                ]
            ],
        },
    }


# NOTE: consider if it would be better to optionally run this function when the attributes
# are already present int he columns.
def add_geo_columns(
    df: gpd.GeoDataFrame,
    geo_columns: list[Literal["bbox", "bounding_quadkey", "quadkey"]],
    quadkey_zoom_level: int = 12,
) -> gpd.GeoDataFrame:
    """
    Augments a GeoDataFrame with bounding box, optional bounding quadkey, and optional quadkey columns.

    Args:
        df: A GeoDataFrame with geometries.
        geo_columns: A list of strings specifying which columns to add. Possible values are 'bbox', 'bounding_quadkey', and 'quadkey'.
        quadkey_zoom_level: Zoom level for quadkey calculation, default is 12.

    Returns:
        A GeoDataFrame possibly with added 'bbox', 'bounding_quadkey', and 'quadkey' columns.
    """
    df = df.copy()
    src_crs = df.crs.to_epsg()
    if src_crs != 4326:
        df = df.to_crs(epsg=4326)

    def calculate_bbox(geom):
        """Calculates the bounding box of a geometry."""
        minx, miny, maxx, maxy = geom.bounds
        return {"xmin": minx, "ymin": miny, "xmax": maxx, "ymax": maxy}

    def calculate_bounding_quadkey(bbox):
        """Calculates the quadkey for the bounding box."""
        return mercantile.quadkey(
            mercantile.bounding_tile(
                bbox["xmin"], bbox["ymin"], bbox["xmax"], bbox["ymax"]
            )
        )

    def get_point_from_geometry(geom):
        """Get a representative point from different geometry types."""
        if geom.geom_type == "Point":
            return geom
        elif geom.geom_type in ["Polygon", "MultiPolygon"]:
            return geom.centroid
        elif geom.geom_type in ["LineString"]:
            return geom.interpolate(0.5, normalized=True)
        # NOTE: The following code is commented out because it is not used in the current implementation.
        # elif geom.geom_type == "MultiPoint":
        #     return geom[0]
        # elif geom.geom_type in ["MultiLineString", "GeometryCollection"]:
        #     return geom[0].interpolate(0.5, normalized=True)
        else:
            msg = f"Unsupported geometry type: {geom.geom_type}"
            raise ValueError(msg)

    # Add bounding box column
    if "bbox" in geo_columns:
        df["bbox"] = df.geometry.apply(calculate_bbox)

    # Add bounding quadkey column
    if "bounding_quadkey" in geo_columns:
        if "bbox" not in df.columns:
            df["bbox"] = df.geometry.apply(calculate_bbox)
        df["bounding_quadkey"] = df["bbox"].apply(calculate_bounding_quadkey)

    # Add quadkey column
    if "quadkey" in geo_columns:
        if quadkey_zoom_level is None:
            msg = (
                "quadkey_zoom_level must be provided when 'quadkey' is in geo_columns."
            )
            raise ValueError(msg)
        if "lon" in df.columns and "lat" in df.columns:
            points = gpd.GeoSeries(
                [Point(xy) for xy in zip(df.lon, df.lat, strict=False)], crs="EPSG:4326"
            )
        else:
            points = df.geometry.apply(get_point_from_geometry).to_crs("EPSG:4326")

        quadkeys = points.apply(
            lambda point: mercantile.quadkey(
                mercantile.tile(point.x, point.y, quadkey_zoom_level)
            )
        )
        df["quadkey"] = quadkeys

    # Revert CRS if necessary
    if src_crs != 4326:
        df = df.to_crs(epsg=src_crs)

    return df



import warnings

import antimeridian
import geopandas as gpd
import pandas as pd
import shapely
from geopandas import GeoDataFrame
from shapely.geometry import LineString, box


def create_buffer_zone(
    gdf: GeoDataFrame,
    planar_crs: int | None = 3857,
    buffer_factor: float = 1.5,
    max_distance: float = 20000,
    use_utm: bool = False,
) -> GeoDataFrame:
    """
    Create a buffer zone around geometries in a GeoDataFrame.

    Args:
        gdf (GeoDataFrame): Input GeoDataFrame with geometries.
        planar_crs (Optional[int]): EPSG code for the planar projection system. Defaults to 3857.
        buffer_factor (float): Factor to multiply the max distance by for the buffer. Defaults to 1.5.
        max_distance (float): Maximum distance for the buffer in meters. Defaults to 20000.
        use_utm (bool): Whether to compute the buffer in the UTM zone of the geometries. Defaults to False.

    Returns:
        GeoDataFrame: GeoDataFrame with the buffer zone geometries.
    """
    src_crs = gdf.crs

    if use_utm:
        planar_crs = gdf.estimate_utm_crs()

    gdf = gdf.to_crs(planar_crs)
    gdf["geometry"] = gdf.buffer(max_distance * buffer_factor)
    gdf = gdf.to_crs(src_crs)

    # TODO:
    # - Add antimeridian handling

    return gdf


def create_antimeridian_buffer(
    max_distance: float, buffer_factor: float = 1.5
) -> gpd.GeoDataFrame:
    """
    Creates a buffered zone around the antimeridian to account for spatial
    operations near the -180/180 longitude line.

    Args:
        max_distance (float): Maximum distance in meters for the buffer around the antimeridian.
        buffer_factor (float, optional): Factor to scale the buffer size. Defaults to 1.5.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame containing the buffered zone around the antimeridian.
    """
    # Create a GeoDataFrame representing the antimeridian line
    antimeridian_line = gpd.GeoDataFrame(
        geometry=[LineString([[-180, -85], [-180, 85]])], crs=4326
    )
    buffer_zone = create_buffer_zone(
        antimeridian_line,
        planar_crs=3857,
        buffer_factor=buffer_factor,
        max_distance=max_distance,
        use_utm=False,
    )

    # Suppress FixWindingWarning from the antimeridian package
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", antimeridian.FixWindingWarning)
        fixed_geometry = shapely.geometry.shape(
            antimeridian.fix_geojson(
                shapely.geometry.mapping(buffer_zone.iloc[0].geometry)
            )
        )

    # Return as GeoDataFrame
    return gpd.GeoDataFrame(geometry=[fixed_geometry], crs=4326)


def add_attributes_from_gdf(
    df: gpd.GeoDataFrame,
    other_gdf: gpd.GeoDataFrame,
    max_distance: float = 20000,
    buffer_factor: float = 1.5,
) -> gpd.GeoDataFrame:
    """
    Adds attributes from a source GeoDataFrame to a target GeoDataFrame based on nearest spatial join.

    Args:
        df (gpd.GeoDataFrame): The target GeoDataFrame to which attributes will be added.
        other_gdf (gpd.GeoDataFrame): The other GeoDataFrame from which attributes will be extracted.
        max_distance (float): The maximum distance for nearest neighbor consideration, in meters.
        buffer_factor (float): Factor to increase the buffer area. Defaults to 1.5.

    Returns:
        gpd.GeoDataFrame: The target GeoDataFrame with added attributes from the source GeoDataFrame.
    """
    # Ensure the transect GeoDataFrame has a point geometry to avoid double intersection.
    transect_origins = gpd.GeoDataFrame(
        df[["transect_id"]],
        geometry=gpd.points_from_xy(df.lon, df.lat, crs=4326),
    )

    antimeridian_buffer = create_antimeridian_buffer(
        max_distance, buffer_factor=buffer_factor
    )

    # Optimization: define the region of interest with a buffer that only works in areas far away from
    # the antimeridian
    if gpd.overlay(transect_origins, antimeridian_buffer).empty:
        roi = gpd.GeoDataFrame(geometry=[box(*transect_origins.total_bounds)], crs=4326)
        roi = gpd.GeoDataFrame(
            geometry=roi.to_crs(3857).buffer(max_distance * 1.5).to_crs(4326)
        )

        # Filter source GeoDataFrame within the region of interest
        other_gdf = gpd.sjoin(other_gdf, roi).drop(columns=["index_right"])

    # Perform nearest neighbor spatial join
    joined = gpd.sjoin_nearest(
        transect_origins.to_crs(3857),
        other_gdf.to_crs(3857),
        max_distance=max_distance,
    ).drop(columns=["index_right", "geometry"])

    # Merge the attributes into the original target GeoDataFrame
    result = pd.merge(df, joined, on="transect_id", how="left").drop_duplicates(
        "transect_id"
    )
    return result


def add_attributes_from_gdfs(
    df: gpd.GeoDataFrame,
    other_gdfs: list[gpd.GeoDataFrame],
    max_distance: float = 20000,
) -> gpd.GeoDataFrame:
    """
    Adds attributes from multiple other GeoDataFrames to a target GeoDataFrame.

    Args:
        df (gpd.GeoDataFrame): The target GeoDataFrame to which attributes will be added.
        other_gdfs (List[gpd.GeoDataFrame]): A list of other GeoDataFrames from which attributes will be extracted.
        max_distance (float): The maximum distance for nearest neighbor consideration, in meters.

    Returns:
        gpd.GeoDataFrame: The target GeoDataFrame with added attributes from all source GeoDataFrames.
    """
    for source_gdf in other_gdfs:
        df = add_attributes_from_gdf(df, source_gdf, max_distance)
    return df



import sys
from pathlib import Path

sys.path.insert(0, str(Path("../../src").absolute()))

project = "coastpy"
copyright = "2022, Floris Calkoen"
author = "Floris Calkoen"

extensions = [
    "myst_parser",  # For Markdown and Jupyter Notebook support
    "autoapi.extension",  # For API documentation
    "sphinx.ext.napoleon",  # For Google style docstrings
    "sphinx.ext.viewcode",  # To include the source code in the docs
    "nbsphinx",  # To include Jupyter Notebooks
    "sphinx_markdown_tables",  # To handle Markdown tables
]

autoapi_dirs = ["../../src"]
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
    ".ipynb": "nbsphinx",
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

nbsphinx_allow_errors = True  # Continue through Jupyter errors

# Additional myst-parser configurations
myst_enable_extensions = [
    "dollarmath",
    "amsmath",
    "deflist",
    "html_image",
]



import logging
import math
import warnings
from typing import Literal

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
import shapely.ops
from pyproj import Transformer, transform
from shapely.geometry import (
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPoint,
    Point,
    base,
)
from shapely.ops import snap, split

from coastpy.utils.dask import silence_shapely_warnings


def shift_point(
    c1: tuple[float, float], c2: tuple[float, float], offset: float
) -> Point:
    """
    Shift a point (`c1`) towards another point (`c2`) by a specified offset distance.
    The new point will lie on the line connecting `c1` and `c2`.

    Args:
        c1 (Tuple[float, float]): The initial point, given as a tuple (x1, y1).
        c2 (Tuple[float, float]): The target point towards which `c1` will be shifted, given as a tuple (x2, y2).
        offset (float): Distance to shift `c1` towards `c2`. If the offset is greater than the distance between
                        `c1` and `c2`, the result will be `c2`.

    Returns:
        Point: A Shapely Point object representing the shifted point.

    Note:
        Adopted from Dirk Eilander at:
        https://svn.oss.deltares.nl/repos/openearthtools/trunk/python/applications/hydrotools/sandbox/osmmodelbuilding/shapely_tools.py
    """
    x1, y1 = c1
    x2, y2 = c2

    # Check for zero length line
    if (x1 - x2) == 0 and (y1 - y2) == 0:
        return Point(x1, y1)

    rel_length = np.minimum(offset / np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2), 1)
    x_new = x1 + (x2 - x1) * rel_length
    y_new = y1 + (y2 - y1) * rel_length

    return Point(x_new, y_new)


def extend_line(line: LineString, offset: float, side: str = "both") -> LineString:
    """
    Extend a LineString in the same orientation on either the start, end, or both sides.

    Args:
        line (LineString): The input line to be extended.
        offset (float): The distance to extend the line by.
        side (str, optional): The side of the line to be extended. It can be "start", "end", or "both".
                              Default is "both".

    Returns:
        LineString: The extended LineString.

    Note:
        Adopted from Dirk Eilander at:
        https://svn.oss.deltares.nl/repos/openearthtools/trunk/python/applications/hydrotools/sandbox/osmmodelbuilding/shapely_tools.py
    """
    sides = ["start", "end"] if side == "both" else [side]

    coords = list(line.coords)
    for s in sides:
        if s == "start":
            p_new = shift_point(coords[0], coords[1], -1.0 * offset)
            coords.insert(0, tuple(p_new.coords[0]))
        elif s == "end":
            p_new = shift_point(coords[-1], coords[-2], -1.0 * offset)
            coords.append(tuple(p_new.coords[0]))
    return LineString(coords)


def extend_transect(line: LineString, length: float, side: str = "both") -> LineString:
    """
    Extend a transect LineString up to a specified total length.

    Args:
        line (LineString): The input transect line to be extended.
        length (float): The total desired length of the transect after extension.
        side (str, optional): The side of the transect to be extended. It can be "start", "end", or "both".
                              Default is "both".

    Returns:
        LineString: The extended LineString transect.

    Raises:
        NotImplementedError: If the desired length is shorter than the current transect length.

    Note:
        This function relies on the `extend_line` function.
    """
    offset = (length - line.length) / 2
    if offset < 0:
        msg = "Transects can only be extended."
        raise NotImplementedError(msg)
    # The extended line contains the original points plus the new points, so we
    # extract the coordinates and use the first and last point to create the new transect.
    line = extend_line(line, offset, side)
    coords = line.coords
    return LineString([coords[0], coords[-1]])


def extract_coordinates(pt: Point | tuple[float, float]) -> tuple[float, float]:
    """Extract coordinates from a Point or tuple.

    Args:
        pt (Union[Point, Tuple[float, float]]): A point object or tuple.

    Returns:
        Tuple[float, float]: The x and y coordinates of the point.
    """
    if isinstance(pt, Point):
        return pt.x, pt.y
    elif (
        isinstance(pt, tuple)
        and len(pt) == 2
        and all(isinstance(coord, (float | int)) for coord in pt)
    ):
        return float(pt[0]), float(pt[1])
    else:
        msg = f"Invalid point provided. Expected a Point object or a tuple of two floats. Received: {pt}"
        raise ValueError(msg)


def extract_endpoints(line: LineString) -> tuple[Point, Point]:
    """Extract the start and end points of a linestring.

    Args:
        line (LineString): A linestring.

    Returns:
        tuple[Point, Point]: The start and end points of the linestring.
    """

    # Check if line is a MultiLineString
    if isinstance(line, MultiLineString):
        msg = "MultiLineString provided instead of LineString."
        raise ValueError(msg)

    # Check validity of LineString
    if not line.is_valid:
        msg = "Invalid LineString provided."
        raise ValueError(msg)

    # Check if LineString has at least two coordinates
    coords = list(line.coords)
    if len(coords) < 2:
        msg = "LineString should have at least two coordinates."
        raise ValueError(msg)

    # Extract start and end points
    start_point = Point(coords[0])
    end_point = Point(coords[-1])

    return start_point, end_point


def get_angle(
    pt1: Point | tuple[float, float], pt2: Point | tuple[float, float]
) -> float:
    """Calculate the angle in degrees between two points.

    The angle is computed in in counter-clockwise direction from the positive x-axis,
    (or east) to the line segment connecting the two points.

    Args:
        pt1 (Union[Point, Tuple[float, float]]): The first point.
        pt2 (Union[Point, Tuple[float, float]]): The second point.

    Returns:
        float: The angle in degrees.
    """
    x1, y1 = extract_coordinates(pt1)
    x2, y2 = extract_coordinates(pt2)
    dy = y2 - y1
    dx = x2 - x1
    return math.degrees(math.atan2(dy, dx))


def get_planar_bearing(
    pt1: Point | tuple[float, float], pt2: Point | tuple[float, float]
) -> float:
    """Calculate the bearing (angle normalized to [0, 360) degrees) between two points.

    The normalization is required to handle cases where the resulting angle is negative,
    which can occur when pt1 is to the right of pt2.

    Args:
        pt1 (Union[Point, Tuple[float, float]]): The first point.
        pt2 (Union[Point, Tuple[float, float]]): The second point.

    Returns:
        float: The bearing in degrees, normalized to [0, 360].

    Example with Points:
        >>> pt1 = Point(1, 1)
        >>> pt2 = Point(2, 2)
        >>> get_bearing(pt1, pt2)
        45.0

    Example with tuples:
        >>> pt1 = (1, 1)
        >>> pt2 = (2, 2)
        >>> get_bearing(pt1, pt2)
        45.0
    """
    angle = get_angle(pt1, pt2)  # angle counter-clock wise from east
    bearing = 90 - angle  # angle to bearing
    return (bearing + 360) % 360


def get_spherical_bearing(
    pt1: tuple[float, float] | Point,
    pt2: tuple[float, float] | Point,
    src_crs: str | int,
    ellps: str = "WGS84",
) -> float:
    """
    Calculate the geodesic bearing between two points.

    Given two points and the source coordinate reference system (CRS),
    this function computes the geodesic bearing between the two points.

    Args:
        pt1 (Union[Tuple[float, float], Point]): The first point.
        pt2 (Union[Tuple[float, float], Point]): The second point.
        src_crs str | int : The source coordinate reference system as an EPSG string or code.

    Returns:
        float: The geodesic bearing in degrees, normalized to [0, 360).

    Example:
        >>> pt1 = (4.9033, 52.3680)
        >>> pt2 = (4.8977, 52.3676)
        >>> src_crs = "EPSG:4326"
        >>> get_geodesic_bearing(pt1, pt2, src_crs)
        295.1521

    Warning:
        This function has not been extensively tested. Use with caution.
    """

    # Issue a user warning
    warnings.warn(
        "This function has not been extensively tested. Use with caution.",
        UserWarning,
        stacklevel=2,
    )

    if not isinstance(pt1, Point):
        pt1 = Point(pt1)
    if not isinstance(pt2, Point):
        pt2 = Point(pt2)

    tf = Transformer.from_crs(src_crs, "EPSG:4326", always_xy=True)
    pt1 = transform(tf.transform, pt1)  # type: ignore
    pt2 = transform(tf.transform, pt2)  # type: ignore

    lon1, lat1 = pt1[0], pt1[1]
    lon2, lat2 = pt2[0], pt2[1]

    geodesic = pyproj.Geod(ellps=ellps)
    fwd_azimuth, _, _ = geodesic.inv(lon1, lat1, lon2, lat2)
    return (fwd_azimuth + 360) % 360


def calculate_point(pt: Point, angle: float, dist: float) -> Point:
    """
    Define a point by distance and angle from a reference point.

    Args:
        pt (Point): The reference point.
        angle (float): The angle in degrees from the reference point.
        dist (float): The distance from the reference point.

    Returns:
        Point: The new point calculated by distance and angle from the reference point.
    """
    bearing = math.radians(angle)
    x = pt.x + dist * math.cos(bearing)
    y = pt.y + dist * math.sin(bearing)
    return Point(x, y)


def split_line_by_points(
    line: LineString, points: MultiPoint, tolerance: float = 0.1
) -> GeometryCollection:
    """
    Split a line by provided points into several segments.

    This function creates segments held in a Shapely GeometryCollection, which is iterable.
    Providing a tolerance is essential due to the float nature of points which might have
    many decimals, causing them not to be found precisely along the line. The snap() function
    from Shapely resolves this by allowing a tolerance.

    Args:
        line (LineString): The linestring to be split.
        points (MultiPoint): Points along the linestring where the line will be split.
        tolerance (float, optional): Tolerance for interpolating points. Defaults to 0.1 meters.

    Returns:
        GeometryCollection: Iterable Shapely line sequence using the `geoms` method.
    """

    return split(snap(line, points, tolerance), points)


def split_line_by_distance(
    line: LineString, distance: float, tolerance: float = 0.1
) -> GeometryCollection:
    """
    Split a line into segments of a specified distance.

    The function ensures global transects are equally spaced. It's especially useful for coastlines that are long and
    cover multiple CRS UTM regions. To describe such coastlines in the local UTM CRS, they should be split into
    several shorter segments.

    The function creates segments held in a Shapely GeometryCollection, which is iterable. It uses a helper function
    `split_line_by_points()` to perform the actual split based on computed points.

    In some cases, like islands that don't have a defined boundary, the last point is extracted directly from the
    list of coordinates. This method is slower, and is used as a fallback.

    Args:
        line (LineString): Coastline described as a Shapely LineString.
        distance (float): Desired segment length in meters.
        tolerance (float, optional): Tolerance for the split operation. Defaults to 0.1 meters.

    Returns:
        GeometryCollection: An iterable collection of linestring segments.
    """

    distances = np.arange(0, line.length, distance)

    # Check for boundary availability, use a fallback if not present
    last_point = (
        line.boundary.geoms[1] if hasattr(line.boundary, "geoms") else line.coords[-1]
    )

    points = MultiPoint([line.interpolate(dist) for dist in distances] + [last_point])

    return split_line_by_points(line, points, tolerance=tolerance)


def merge_lines(df: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """
    Merge lines in the GeoDataFrame into a continuous line.

    This function takes in a GeoDataFrame containing line geometries, and
    attempts to merge them into a continuous line using shapely's linemerge
    operation.

    Args:
        df (GeoDataFrame): Input GeoDataFrame with line geometries.

    Returns:
        GeoDataFrame: GeoDataFrame containing the merged line.

    Note:
        If the input dataframe is empty, an empty GeoDataFrame is returned.
    """

    # Check for an empty dataframe
    if df.empty:
        return gpd.GeoDataFrame(geometry=[], crs=df.crs)

    merged_line = shapely.ops.linemerge(
        df.explode(index_parts=False).geometry.to_list()
    )

    return (
        gpd.GeoDataFrame(geometry=[merged_line], crs=df.crs)
        .explode(index_parts=False)
        .reset_index(drop=True)
    )


def overlay_by_grid(df: gpd.GeoDataFrame, grid: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """
    Overlay a GeoDataFrame with a given grid and return the intersected geometries.

    Args:
        df (gpd.GeoDataFrame): Input GeoDataFrame to be overlaid.
        grid (gpd.GeoDataFrame): The grid used for overlay.

    Returns:
        gpd.GeoDataFrame: Overlaid GeoDataFrame with intersected geometries.
    """

    return gpd.overlay(
        df,
        grid,
        keep_geom_type=False,
    ).explode(column="geometry", index_parts=False)


def generate_offset_line(line: LineString, offset: float) -> LineString:
    """
    Generate an offset line from the original line at a specified distance using offset_curve method.

    Args:
        line (LineString): The original line from which the offset is generated.
        offset (float): The distance for the offset. Positive values offset to the left,
            and negative values offset to the right.

    Returns:
        LineString: The offset line generated from the original line.
    """
    return line.offset_curve(offset) if offset != 0 else line


def get_rotation_angle(
    pt1: Point | tuple[float, float],
    pt2: Point | tuple[float, float],
    target_axis: Literal[
        "closest", "vertical", "horizontal", "horizontal-right-aligned"
    ] = "closest",
) -> float | None:
    """
    Computes the correct rotation angle to align with a specified axis.

    Args:
        pt1 (Union[Point, Tuple[float, float]]): The starting point of the transect.
        pt2 (Union[Point, Tuple[float, float]]): The ending point of the transect.
        target_axis (Literal["closest", "vertical", "horizontal", "horizontal-right-aligned"], optional):
            The target axis to align the transect. Defaults to "closest".

    Returns:
        float: The rotation angle in degrees. Positive values represent counterclockwise rotation.

    Raises:
        ValueError: If an invalid target axis is provided or if the bearing is out of the expected range.
    """

    x1, y1 = extract_coordinates(pt1)
    x2, y2 = extract_coordinates(pt2)

    angle = get_angle(pt1, pt2)
    bearing = get_planar_bearing(pt1, pt2)
    logging.info(f"Angle between points: {angle} degrees.")
    logging.info(f"Bearing between points: {bearing} degrees.")

    if x1 == x2 or y1 == y2:
        return 0

    if target_axis == "closest":
        angle_rotations = {
            (0, 45): lambda b: b,
            (45, 90): lambda b: -(90 - b),
            (90, 135): lambda b: b - 90,
            (135, 180): lambda b: -(180 - b),
            (180, 225): lambda b: b - 180,
            (225, 270): lambda b: -(270 - b),
            (270, 315): lambda b: b - 270,
            (315, 360): lambda b: -(360 - b),
        }

    elif target_axis == "horizontal":
        angle_rotations = {
            (0, 90): lambda b: -(90 - b),
            (90, 180): lambda b: b - 90,
            (180, 270): lambda b: -(270 - b),
            (270, 360): lambda b: b - 270,
        }

    # TODO: rename to landward right
    elif target_axis == "horizontal-right-aligned":
        angle_rotations = {
            (0, 90): lambda b: 90 + b,
            (90, 180): lambda b: -(270 - b),
            (180, 270): lambda b: -(270 - b),
            (270, 360): lambda b: b - 270,
        }

    elif target_axis == "vertical":
        angle_rotations = {
            (0, 90): lambda b: b,
            (90, 180): lambda b: b,
            (180, 270): lambda b: -(270 - b),
            (270, 360): lambda b: -(360 - b),
        }

    else:
        msg = f"Invalid target_axis: {target_axis}. Must be one of 'closest', 'vertical', 'horizontal', or 'horizontal-right-aligned'."
        raise ValueError(msg)

    for (lower_bound, upper_bound), rotation_func in angle_rotations.items():
        if lower_bound <= bearing < upper_bound:
            return rotation_func(bearing)

    msg = "Invalid bearing computed. Expected bearing within range [0, 360]."
    raise ValueError(msg)


def crosses_antimeridian(df: gpd.GeoDataFrame) -> pd.Series:
    """
    Determines whether LineStrings in a GeoDataFrame cross the International Date Line.

    Args:
        df (gpd.GeoDataFrame): Input GeoDataFrame with LineString geometries.

    Returns:
        pd.Series: Series indicating whether each LineString crosses the antimeridian.

    Example:
        >>> df = gpd.read_file('path_to_file.geojson')
        >>> df['crosses_antimeridian'] = crosses_antimeridian(df)
        >>> print(df['crosses_antimeridian'])
    """
    # Ensure the CRS is in degrees (longitude, latitude)
    if df.crs.to_epsg() != 4326:
        df = df.to_crs(4326)

    # Extract coordinates from the geometry
    coords = df.geometry.apply(lambda geom: np.array(geom.coords.xy).T)

    # Vectorized check for antimeridian crossing
    def crosses(coords: np.ndarray) -> bool:
        # Calculate differences between consecutive longitudes
        longitudes = coords[:, 0]
        lon_diff = np.diff(longitudes)

        # Check if the difference is greater than 180 degrees (indicating a crossing)
        crosses = np.abs(lon_diff) > 180
        return np.any(crosses)

    # Apply the vectorized check across all geometries
    return coords.apply(crosses)


def _buffer_geometry(
    geom: base.BaseGeometry, src_crs: str | int, buffer_dist: float
) -> base.BaseGeometry:
    """
    Buffers a single geometry in its appropriate UTM projection and reprojects it back to the original CRS.

    Args:
        geom (shapely.geometry.base.BaseGeometry): The geometry to buffer.
        src_crs (Union[str, int]): The original CRS of the geometry.
        buffer_dist (float): The buffer distance in meters.

    Returns:
        base.BaseGeometry: The buffered geometry in the original CRS.
    """
    # Estimate the UTM CRS based on the geometry's location
    utm_crs = gpd.GeoSeries([geom], crs=src_crs).estimate_utm_crs()

    # Reproject the geometry to UTM, apply the buffer, and reproject back to the original CRS
    geom_utm = gpd.GeoSeries([geom], crs=src_crs).to_crs(utm_crs).iloc[0]
    buffered_utm = geom_utm.buffer(buffer_dist)
    buffered_geom = gpd.GeoSeries([buffered_utm], crs=utm_crs).to_crs(src_crs).iloc[0]

    return buffered_geom


def buffer_geometries_in_utm(
    geo_data: gpd.GeoSeries | gpd.GeoDataFrame, buffer_dist: float
) -> gpd.GeoSeries | gpd.GeoDataFrame:
    """
    Buffer all geometries in a GeoSeries or GeoDataFrame in their appropriate UTM projections and return
    the buffered geometries in the original CRS.

    Args:
        geo_data (Union[gpd.GeoSeries, gpd.GeoDataFrame]): Input GeoSeries or GeoDataFrame containing geometries.
        buffer_dist (float): Buffer distance in meters.

    Returns:
        Union[gpd.GeoSeries, gpd.GeoDataFrame]: Buffered geometries in the original CRS.
    """
    # Determine if the input is a GeoDataFrame or a GeoSeries
    is_geodataframe = isinstance(geo_data, gpd.GeoDataFrame)

    # Extract the geometry series from the GeoDataFrame, if necessary
    geom_series = geo_data.geometry if is_geodataframe else geo_data

    # Ensure the input data has a defined CRS
    if geom_series.crs is None:
        msg = "Input GeoSeries or GeoDataFrame must have a defined CRS."
        raise ValueError(msg)

    # Buffer each geometry using the UTM projection and return to original CRS
    buffered_geoms = geom_series.apply(
        lambda geom: _buffer_geometry(geom, geom_series.crs, buffer_dist)
    )

    # Return the modified GeoDataFrame or GeoSeries with the buffered geometries
    if is_geodataframe:
        geo_data = geo_data.assign(geometry=buffered_geoms)
        return geo_data
    else:
        return buffered_geoms


def add_line_length(
    gdf: gpd.GeoDataFrame, dst_crs: str | int, length_col: str
) -> gpd.GeoDataFrame:
    """
    Adds a column to the GeoDataFrame `gdf` with the length of each LineString in meters,
    in the target CRS `target_crs`. The column will be named `length_col`.

    Parameters:
    -----------
    gdf : GeoDataFrame
        The input GeoDataFrame containing LineString geometries.
    target_crs : str
        The target CRS to convert the geometries to, in EPSG format (e.g. 'EPSG:4326', 4326).
    length_col : str
        The name of the new column to add to the GeoDataFrame.

    Returns:
    --------
    GeoDataFrame
        The input GeoDataFrame with the new length column added.
    """
    src_crs = gdf.crs
    gdf = gdf.to_crs(dst_crs)
    gdf[length_col] = gdf.geometry.length
    gdf = gdf.to_crs(src_crs)
    return gdf


def get_xy_range(
    gdf: gpd.GeoDataFrame,
) -> tuple[tuple[float, float], tuple[float, float]]:
    """
    Extracts the min and max longitude (x) and latitude (y) from a GeoDataFrame
    and returns them as a formatted xy range suitable for geoviews.

    Args:
        gdf (gpd.GeoDataFrame): The input GeoDataFrame.

    Returns:
        Tuple[Tuple[float, float], Tuple[float, float]]:
            The x range (min lon, max lon) and y range (min lat, max lat).
    """

    x_range = tuple(gdf.total_bounds[[0, 2]])
    y_range = tuple(gdf.total_bounds[[1, 3]])
    return x_range, y_range


def add_geometry_lengths(
    df: gpd.GeoDataFrame,
    crs: str,
    group_col: str,
    geometry_length_col: str = "geometry_length",
    total_length_col: str = "partial_FID_length",
) -> pd.DataFrame:
    """
    Computes and adds geometry length for a given GeoDataFrame in a specified CRS.

    Args:
        df (gpd.GeoDataFrame): The input GeoDataFrame.
        crs (str): The CRS used for computing the length.
        group_col (str, optional): Column used for grouping and calculating total lengths.
        geometry_length_col (str, optional): Name of the column where individual geometry lengths will be stored. Defaults to 'geometry_length'.
        total_length_col (str, optional): Name of the column where total lengths per group will be stored. Defaults to 'partial_FID_length'.

    Returns:
        pd.DataFrame: A DataFrame with added length columns.

    Warnings:
        This function might provide inaccurate total lengths for 'FID's that extend across multiple CRSs. It is recommended to overlay the input data with a UTM grid and group it by the appropriate CRS before computing lengths using this function.
    """

    silence_shapely_warnings()

    # Warning about potential inaccuracies
    warnings.warn(
        (
            "This function might provide inaccurate total lengths for 'FID's that"
            " extend across multiple CRSs. It is recommended to overlay the input data"
            " with a UTM grid and group it by the appropriate CRS before computing"
            " lengths using this function."
        ),
        UserWarning,
        stacklevel=2,
    )

    # Compute geometry length in the specified CRS
    df_temp = df.to_crs(crs)
    df_temp[geometry_length_col] = df_temp.geometry.length
    df = df_temp.to_crs(df.crs)

    # Compute total length per group
    total_lengths = (
        df.groupby(group_col)[geometry_length_col]
        .sum()
        .rename(total_length_col)
        .reset_index()
    )

    # Merge to the main DataFrame and drop temporary geometry_length column
    df = pd.merge(df.drop(columns=[geometry_length_col]), total_lengths, on=group_col)

    return df



from dataclasses import dataclass, field

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
import shapely
from pyproj import Transformer
from shapely import LineString, Point, wkt
from shapely.ops import transform

from coastpy.geo.ops import (
    calculate_point,
    extract_coordinates,
    extract_endpoints,
    get_angle,
    get_planar_bearing,
)


@dataclass
class Transect:
    """Dataclass for transects"""

    transect_id: str
    tr_origin: Point
    tr_length: int
    tr_angle: float
    utm_epsg: int | str
    dst_crs: int | str

    _geometry: LineString = field(init=False, repr=False)
    _lon: float = field(init=False, repr=False)
    _lat: float = field(init=False, repr=False)
    _bearing: float = field(init=False, repr=False)

    @property
    # @lru_cache(maxsize=1)  # does not work when using dask distributed
    def geometry(self):
        utm_crs_epsg = pyproj.CRS.from_user_input(self.utm_epsg).to_epsg()
        dst_crs_epsg = pyproj.CRS.from_user_input(self.dst_crs).to_epsg()

        pt1 = calculate_point(self.tr_origin, self.tr_angle + 90, 0.5 * self.tr_length)
        opposite_angle = get_angle(pt1, self.tr_origin)
        pt2 = calculate_point(pt1, opposite_angle, self.tr_length)
        self._bearing = get_planar_bearing(pt1, pt2)

        tf = Transformer.from_crs(utm_crs_epsg, dst_crs_epsg, always_xy=True)
        geometry = transform(tf.transform, LineString([pt1, pt2]))

        tf_4326 = Transformer.from_crs(utm_crs_epsg, 4326, always_xy=True)
        lon, lat = next(iter(transform(tf_4326.transform, self.tr_origin).coords))
        tr_origin = transform(tf.transform, self.tr_origin)

        self._geometry = geometry
        self._tr_origin = tr_origin
        self._lon = lon
        self._lat = lat

        return geometry

    def to_dict(self):
        return {
            "transect_id": self.transect_id,
            "geometry": self.geometry,
            "lon": self._lon,
            "lat": self._lat,
            "tr_origin": wkt.dumps(self._tr_origin),
            "bearing": self._bearing,
            "utm_epsg": self.utm_epsg,
            "src_crs": self.dst_crs,
        }

    def __hash__(self):
        return hash(
            (
                self.transect_id,
                self.tr_origin,
                self.tr_length,
                self.utm_epsg,
                self.dst_crs,
            )
        )


def transect_geometry(
    start_datum: Point,
    tr_origin: Point,
    end_datum: Point,
    transect_length: float,
    rot_anti: np.ndarray,
    rot_clock: np.ndarray,
) -> LineString:
    """
    Generate a transect based on the input coastline segment and transect length.

    Args:
        start_datum (Point): Start point of the coastal segment.
        tr_origin (Point): Middle point of the coastal segment (transect origin).
        end_datum (Point): End point of the coastal segment.
        transect_length (float): Transect length in meters.
        rot_anti (np.ndarray): Anti-clockwise rotation matrix.
        rot_clock (np.ndarray): Clockwise rotation matrix.

    Returns:
        LineString: Transect as Shapely LineString.

    Example:
        >>> start = Point(0, 0)
        >>> mid = Point(1, 1)
        >>> end = Point(2, 2)
        >>> transect_length = 1
        >>> ROT_ANTI = np.array([[0, -1], [1, 0]])
        >>> ROT_CLOCK = np.array([[0, 1], [-1, 0]])
        >>> transect_geometry(start, mid, end, transect_length, ROT_ANTI, ROT_CLOCK)
        <shapely.geometry.linestring.LineString object at ...>
    """

    transect_length = transect_length / 2

    segment_vector = np.array(
        [[end_datum.x - start_datum.x], [end_datum.y - start_datum.y]]
    )
    perpendicular_anti = np.dot(rot_anti, segment_vector)
    perpendicular_clock = np.dot(rot_clock, segment_vector)

    # Normalize and scale the perpendicular vectors
    perpendicular_anti = (
        perpendicular_anti / np.linalg.norm(perpendicular_anti)
    ) * transect_length
    perpendicular_clock = (
        perpendicular_clock / np.linalg.norm(perpendicular_clock)
    ) * transect_length

    transect_start = (
        tr_origin.x + float(perpendicular_anti[0]),
        tr_origin.y + float(perpendicular_anti[1]),
    )
    transect_end = (
        tr_origin.x + float(perpendicular_clock[0]),
        tr_origin.y + float(perpendicular_clock[1]),
    )

    return LineString([transect_start, transect_end])


def make_transect_origins(
    line: shapely.geometry.LineString, spacing: float
) -> list[float]:
    """
    Generate transect origins along a LineString with a given spacing.

    Args:
        line: LineString representing the transect line.
        spacing: Distance between transect origins.

    Returns:
        List of transect origins as distances from the start of the LineString.
    """
    n_transects = int(line.length / spacing)
    leftover = line.length % spacing

    if leftover < spacing * 0.5:
        n_transects = n_transects - 1

    if n_transects == 0:
        return [line.length / 2]

    start = (line.length - n_transects * spacing) * 0.5
    end = start + n_transects * spacing
    points = np.arange(start, end, spacing)
    return points.tolist()


def make_transects(
    coastline: gpd.GeoSeries,
    osm_coastline_id: str,
    src_crs: str,
    utm_epsg: str,
    dst_crs: str,
    spacing: float,
    transect_length: int = 2000,
    smooth_distance: float = 1.0e-3,
) -> gpd.GeoDataFrame:
    """
    Generate transects along a coastline.

    Args:
        coastline: GeoSeries representing the coastline.
        osm_coastline_id: Name of the coastline.
        src_crs: Source CRS of the coastline.
        utm_epsg: UTM CRS for local coordinate transformation.
        dst_crs: Destination CRS for the transects.
        spacing: Distance between transects.
        idx: Index of the function call.
        transect_length: Length of the transects.
        smooth_distance: Distance used to smooth the coastline.

    Returns:
        GeoDataFrame containing transects with their attributes and geometry.
    """
    # Convert coastline to local UTM CRS
    tf = Transformer.from_crs(src_crs, utm_epsg, always_xy=True)
    coastline = transform(tf.transform, coastline)

    origins = make_transect_origins(coastline, spacing)

    transects = []
    for origin in origins:
        # Derive transect angle from nearby points of transect origin
        tr_origin = coastline.interpolate(origin)
        pt1 = coastline.interpolate(origin - smooth_distance)
        pt2 = coastline.interpolate(origin + smooth_distance)

        angle = get_angle(pt1, pt2)
        tr_id = f"cl{int(osm_coastline_id)}tr{int(origin)}"

        tr = Transect(
            transect_id=tr_id,
            tr_angle=angle,
            tr_origin=tr_origin,
            tr_length=transect_length,
            utm_epsg=utm_epsg,
            dst_crs=dst_crs,
        )

        transects.append(tr.to_dict())

    if not transects:
        return None

    # when pyarrow is more stable, use pyarrow dtypes instead
    # column_datatypes = {
    #     "transect_id": "string[pyarrow]",
    #     "lon": "float64[pyarrow]",
    #     "lat": "float64[pyarrow]",
    #     "tr_origin": "string[pyarrow]",
    #     "bearing": "float64[pyarrow]",
    #     "utm_epsg": "int32[pyarrow]",
    #     "src_crs": "int32[pyarrow]",
    #     "osm_coastline_id": "int32[pyarrow]",
    # }

    column_datatypes = {
        "transect_id": "string",
        "lon": "float64",
        "lat": "float64",
        "tr_origin": "string",
        "bearing": "float64",
        "utm_epsg": "int32",
        "src_crs": "int32",
        "osm_coastline_id": "int32",
    }

    # TODO: instead of dropping transects that cross date line, create MultiLinestrings?
    return (
        gpd.GeoDataFrame(transects, geometry="geometry", crs=dst_crs)
        .reset_index(drop=True)
        # .pipe(drop_transects_crossing_antimeridian)
        .assign(osm_coastline_id=osm_coastline_id)
        .astype(column_datatypes)
    )


def generate_transects_from_coastline(
    coastline: LineString,
    transect_length: float,
    spacing: float | int,
    osm_coastline_id: int,
    osm_coastline_is_closed: bool,
    osm_coastline_length: int,
    src_crs: str | int,
    utm_epsg: str | int,
    dst_crs: str | int,
    smooth_distance: float = 1e-3,
) -> gpd.GeoDataFrame:
    """
    Generate a list of transects from a coastline geometry.

    Args:
        coastline (LineString): The coastline geometry.
        transect_length (float): Length of the transects.
        osm_coastline_id (int): ID for the coastline.
        osm_coastline_is_closed (bool): If the source OSM coastline is closed.
        osm_coastline_length (int): length of the coastline.
        src_crs (str): Source CRS of the coastline geometry.
        utm_epsg (str): UTM CRS for local coordinate transformation.
        dst_crs (str): Target CRS for the transects.
        smooth_distance (float, optional): Smoothing distance. Defaults to 1e-3.

    Returns:
        gpd.GeoDataFrame: A GeoDataFrame with transects and related attributes.
    """
    # Define a template empty geodataframe with specified datatypes
    META = gpd.GeoDataFrame(
        {
            "transect_id": pd.Series([], dtype="string"),
            "lon": pd.Series([], dtype="float32"),
            "lat": pd.Series([], dtype="float32"),
            "bearing": pd.Series([], dtype="float32"),
            "utm_epsg": pd.Series([], dtype="int32"),
            # NOTE: leave here because before we used to store the coastline name
            # "osm_coastline_id": pd.Series([], dtype="string"),
            "osm_coastline_is_closed": pd.Series([], dtype="bool"),
            "osm_coastline_length": pd.Series([], dtype="int32"),
            "geometry": gpd.GeoSeries([], dtype="geometry"),
        },
        crs=dst_crs,
    )

    ROT_ANTI = np.array([[0, -1], [1, 0]])
    ROT_CLOCK = np.array([[0, 1], [-1, 0]])

    dtypes = META.dtypes.to_dict()
    column_order = META.columns.to_list()

    tf = Transformer.from_crs(src_crs, utm_epsg, always_xy=True)
    coastline = transform(tf.transform, coastline)

    origins = make_transect_origins(coastline, spacing)
    tr_origins = [coastline.interpolate(o) for o in origins]

    if not tr_origins:
        return META

    start_datums = [coastline.interpolate(o - smooth_distance) for o in origins]
    end_datums = [coastline.interpolate(o + smooth_distance) for o in origins]

    transects = [
        transect_geometry(start, origin, end, transect_length, ROT_ANTI, ROT_CLOCK)
        for start, origin, end in zip(start_datums, tr_origins, end_datums, strict=True)
    ]

    endpoints = [extract_endpoints(tr) for tr in transects]
    pt1s, pt2s = zip(*endpoints, strict=False)
    bearings = [
        get_planar_bearing(pt1, pt2) for pt1, pt2 in zip(pt1s, pt2s, strict=True)
    ]

    tf_4326 = Transformer.from_crs(utm_epsg, 4326, always_xy=True)
    tr_origins_4326 = [
        transform(tf_4326.transform, tr_origin) for tr_origin in tr_origins
    ]

    lons, lats = zip(*[extract_coordinates(p) for p in tr_origins_4326], strict=True)

    transect_ids = [f"{osm_coastline_id}tr{int(o)}" for o in origins]

    return (
        gpd.GeoDataFrame(
            {
                "transect_id": transect_ids,
                "lon": lons,
                "lat": lats,
                "bearing": bearings,
                "geometry": transects,
            },
            crs=utm_epsg,
        )
        .to_crs(dst_crs)
        # NOTE: leave here because before we used to store the coastline name
        # .assign(osm_coastline_id=osm_coastline_id)
        .assign(utm_epsg=utm_epsg)
        .assign(osm_coastline_is_closed=osm_coastline_is_closed)
        .assign(osm_coastline_length=osm_coastline_length)
        .loc[:, column_order]
        .astype(dtypes)
    )



import logging
import os
import pathlib
from enum import Enum, auto
from sysconfig import get_python_version


class ComputeInstance(Enum):
    LOCAL = auto()
    SLURM = auto()


def is_interactive() -> bool:
    """
    Check if the code is running in a Jupyter Notebook environment.
    """
    try:
        shell = get_python_version().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter Notebook or JupyterLab
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal or IPython cognsole
        else:
            return False  # Other interactive shells
    except NameError:
        return False  # Not in an interactive shell


def get_proj_dir(proj_name) -> pathlib.Path:
    """
    Get the project directory path.

    Returns:
        A `pathlib.Path` object representing the project directory path.
    """

    def resolve_proj_path(cwd: pathlib.Path, proj_name: str) -> pathlib.Path:
        root = cwd.root
        proj_dir = cwd
        while proj_dir.name != proj_name:
            proj_dir = proj_dir.parent
            if str(proj_dir) == root:
                msg = "Reached root depth - cannot resolve project path."
                raise ValueError(msg)

        return proj_dir

    cwd = pathlib.Path().resolve() if is_interactive() else pathlib.Path(__file__)

    proj_dir = resolve_proj_path(cwd, proj_name=proj_name)
    return proj_dir


def detect_instance_type() -> ComputeInstance:
    """Detect if running on a local machine or a remote SLURM cluster."""
    if (
        "SLURM_JOB_ID" in os.environ
        or "SLURM_NODELIST" in os.environ
        or "SLURM_JOB_NAME" in os.environ
    ):
        return ComputeInstance.SLURM
    else:
        return ComputeInstance.LOCAL


def configure_instance() -> ComputeInstance:
    """Configure the instance based on its type.

    This function detects whether the code is running on a local machine or a SLURM cluster.
    It logs the type of instance and returns the detected instance type. Use this function
    to add logic based on the compute environment.

    Returns:
        ComputeInstance: The type of the compute instance.

    Raises:
        ValueError: If the compute instance type is unknown.
    """
    instance = detect_instance_type()
    if instance == ComputeInstance.SLURM:
        logging.info("Running on a SLURM cluster.")
        return instance
    elif instance == ComputeInstance.LOCAL:
        logging.info("Running on a local machine.")
        return instance
    else:
        msg = "Unknown compute instance type."
        raise ValueError(msg)



from .config import configure_instance, detect_instance_type
from .dask import DaskClientManager
from .size import readable_bytes, size_to_bytes

__all__ = [
    "detect_instance_type",
    "configure_instance",
    "DaskClientManager",
    "size_to_bytes",
    "readable_bytes",
]



import logging
from typing import Any

import dask
from distributed import Client

from coastpy.utils.config import ComputeInstance


class DaskClientManager:
    """Manager for creating Dask clients based on compute instance type.

    This class supports the creation of local and SLURM Dask clusters,
    with optional configuration from external files.

    Attributes:
        config_path (Optional[str]): Path to a Dask configuration file.
    """

    def __init__(self):
        """Initialize the DaskClientManager, optionally loading a config file.

        Args:
            config_path (Optional[str]): Path to the configuration file.
        """
        dask.config.refresh()

    def create_client(self, instance_type: ComputeInstance, *args: Any, **kwargs: Any):
        """Create a Dask client based on the instance type.

        Args:
            instance_type (ComputeInstance): The type of the compute instance.
            *args: Additional positional arguments for client creation.
            **kwargs: Additional keyword arguments for client creation.

        Returns:
            Client: The Dask client.

        Raises:
            ValueError: If the instance type is not recognized.
        """

        if instance_type.name == "LOCAL":
            return self._create_local_client(*args, **kwargs)
        elif instance_type.name == "SLURM":
            return self._create_slurm_client(*args, **kwargs)
        else:
            msg = "Unknown compute instance type."
            raise ValueError(msg)

    def _create_local_client(self, *args: Any, **kwargs: Any) -> Client:
        """Create a local Dask client with potential overrides.

        Args:
            *args: Additional positional arguments for client creation.
            **kwargs: Additional keyword arguments for client creation.

        Returns:
            Client: The Dask local client.
        """
        # Set default values
        from distributed import Client

        configs = {
            "threads_per_worker": 1,
            "processes": True,
            "n_workers": 5,
            "local_directory": "/tmp",
        }

        # Update defaults with any overrides provided in kwargs
        configs.update(kwargs)

        # Create and return the Dask Client using the updated parameters
        return Client(*args, **configs)

    def _create_slurm_client(self, *args: Any, **kwargs: Any) -> Client:
        """Create a SLURM Dask client with potential overrides.

        Args:
            *args: Additional positional arguments for client creation.
            **kwargs: Additional keyword arguments for client creation.

        Returns:
            Client: The Dask SLURM client.
        """

        try:
            from dask_jobqueue import SLURMCluster
        except ImportError as e:
            msg = (
                "dask_jobqueue is required to create a SLURM Dask client. "
                "Please install it using `pip install dask-jobqueue`."
            )
            raise ImportError(msg) from e

        slurm_configs = {
            "cores": 10,  # Cores per worker
            "processes": 10,  # Processes per worker
            # "n_workers": 10,
            "memory": "120GB",  # Memory per worker
            # "local_directory": "/scratch/frcalkoen/tmp",
            "walltime": "3:00:00",
            "log_directory": "/scratch/frcalkoen/tmp",
        }
        # Update default values with any overrides provided in kwargs
        slurm_configs.update(kwargs)

        # Create the SLURM cluster
        cluster = SLURMCluster(*args, **slurm_configs)
        cluster.scale(jobs=2)

        logging.info(f"{cluster.job_script()}")

        # cluster.scale(jobs=5)

        # min_jobs = kwargs.pop(
        #     "minimum_jobs", dask.config.get("jobqueue.adaptive.minimum", 1)
        # )
        # max_jobs = kwargs.pop(
        #     "maximum_jobs", dask.config.get("jobqueue.adaptive.maximum", 30)
        # )

        # cluster.adapt(minimum_jobs=min_jobs, maximum_jobs=max_jobs)
        return Client(cluster)

    # def _create_slurm_client(self, *args: Any, **kwargs: Any):
    #     """Create a SLURM Dask client with potential overrides.

    #     Args:
    #         *args: Additional positional arguments for client creation.
    #         **kwargs: Additional keyword arguments for client creation.

    #     Returns:
    #         Client: The Dask SLURM client.
    #     """
    #     from dask_jobqueue import SLURMCluster

    #     min_jobs = kwargs.pop(
    #         "minimum_jobs", dask.config.get("jobqueue.adaptive.minimum", 1)
    #     )
    #     max_jobs = kwargs.pop(
    #         "maximum_jobs", dask.config.get("jobqueue.adaptive.maximum", 30)
    #     )

    #     cluster = SLURMCluster(*args, **kwargs)
    #     cluster.adapt(minimum_jobs=min_jobs, maximum_jobs=max_jobs)
    #     return cluster.get_client()


def silence_shapely_warnings() -> None:
    """Suppress specific warnings commonly encountered in Shapely geometry operations."""
    import warnings

    warnings_to_ignore: list[str] = [
        "invalid value encountered in buffer",
        "invalid value encountered in intersection",
        "invalid value encountered in unary_union",
    ]

    for warning in warnings_to_ignore:
        warnings.filterwarnings("ignore", message=warning)



import logging
import pathlib
from typing import Any

import fsspec
import geopandas as gpd
import pandas as pd
import xarray as xr

from coastpy.io.utils import is_local_file_path


def to_https_url(href: str, storage_options: dict | None = None) -> str:
    """Converts a cloud storage href to its corresponding HTTPS URL.

    Args:
        href (str): The href string to be converted.
        storage_options (dict, optional): Dictionary of storage options that can contain
            platform-specific settings. For Azure Blob Storage, it should contain
            the 'account_name' key.

    Returns:
        str: The corresponding HTTPS URL.

    Raises:
        ValueError: If the protocol in href is unknown or unsupported or if required
            options are missing in 'storage_options'.
    """

    # Local file system
    if storage_options is None:
        storage_options = {}
    if pathlib.Path(href).exists():
        return f"file://{href}"

    # Google Cloud Storage
    for protocol in ["gs://", "gcs://"]:
        if href.startswith(protocol):
            return href.replace(protocol, "https://storage.googleapis.com/")

    # Azure Blob Storage
    if href.startswith(("az://", "azure://", "abfs://")):
        if "account_name" not in storage_options:
            msg = (
                "For Azure Blob Storage hrefs, the 'account_name' is required in"
                " 'storage_options'."
            )
            raise ValueError(msg)

        _, path = href.split("://", 1)
        account_name = storage_options["account_name"]
        return f"https://{account_name}.blob.core.windows.net/{path}"

    # AWS S3
    if href.startswith("s3://"):
        _, rest = href.split("://", 1)
        bucket, key = rest.split("/", 1)
        return f"https://{bucket}.s3.amazonaws.com/{key}"

    msg = f"Unknown or unsupported protocol in href: {href}"
    raise ValueError(msg)


def to_uri(href_or_uri: str, protocol: str | None = None) -> str:
    """
    Converts an HTTPS URL to its corresponding cloud storage URI based on the specified protocol.
    If the input is already a valid cloud URI, it adjusts the URI based on the provided protocol or returns it as is.

    Args:
        href_or_uri (str): The HTTPS URL or cloud URI to be converted or validated.
        protocol (Optional[str]): The desired cloud protocol ("gs", "gcs", "az", "azure", "abfs", "s3").

    Returns:
        str: The corresponding URI in the desired protocol.

    Raises:
        ValueError: If the protocol is unknown, unsupported, or href_or_uri does not match the expected format.

    Example:
        >>> to_uri_protocol("https://storage.googleapis.com/my-bucket/my-file.txt", protocol="gs")
        'gs://my-bucket/my-file.txt'
    """
    valid_protocols = ["gs", "gcs", "az", "azure", "abfs", "s3"]
    href_or_uri = str(href_or_uri)

    # Google Cloud Storage
    gcs_formats = [
        "https://storage.googleapis.com/",
        "https://storage.cloud.google.com/",
    ]

    for fmt in gcs_formats:
        if href_or_uri.startswith(fmt):
            prefix = href_or_uri.replace(fmt, "")
            if protocol in [None, "gs"]:
                return f"gs://{prefix}"
            elif protocol == "gcs":
                return f"gcs://{prefix}"
            else:
                msg = (
                    "For Google Cloud Storage href, valid protocols are 'gs' and"
                    f" 'gcs', but got '{protocol}'."
                )
                raise ValueError(msg)

    # Azure Blob Storage
    if ".blob.core.windows.net/" in href_or_uri:
        # Split the href to get the container and path
        container_and_path = href_or_uri.split(".blob.core.windows.net/", 1)[1]
        if protocol in [None, "az", "azure"]:
            return f"az://{container_and_path}"
        elif protocol == "abfs":
            return f"abfs://{container_and_path}"
        msg = (
            "For Azure Blob Storage href, valid protocols are 'az', 'azure', and"
            f" 'abfs', but got '{protocol}'."
        )
        raise ValueError(msg)

    # AWS S3
    if href_or_uri.startswith("https://s3.") and ".amazonaws.com/" in href_or_uri:
        parts = href_or_uri.replace("https://", "").split(".amazonaws.com/")
        bucket = parts[0].replace("s3.", "")
        key = parts[1]
        if protocol in [None, "s3"]:
            return f"s3://{bucket}/{key}"
        msg = f"For AWS S3 href, the valid protocol is 's3', but got '{protocol}'."
        raise ValueError(msg)

    # If the input is already a valid cloud URI
    for uri_prefix in ["gs://", "gcs://", "az://", "azure://", "abfs://", "s3://"]:
        if href_or_uri.startswith(uri_prefix):
            if protocol is None:
                return href_or_uri
            else:
                return to_uri(href_or_uri.replace(uri_prefix, "https://", 1), protocol)

    if is_local_file_path(href_or_uri):
        return href_or_uri

    msg = (
        f"Unknown or unsupported protocol: {protocol}. Valid protocols are"
        f" {', '.join(valid_protocols)}.\nOr href_or_uri does not match expected"
        f" format: {href_or_uri}"
    )

    raise ValueError(msg)


def write_block(
    block: xr.DataArray,
    blob_name: str,
    storage_options: dict[str, Any] | None = None,
    profile_options: dict[str, Any] | None = None,
    overwrite: bool = True,
) -> int:
    """
    Write a given block of data to a specified cloud storage location.

    Args:
        block (xr.DataArray): The data block to be written to cloud storage.
        blob_name (str): The target storage path, including the blob's name.
        storage_options (Optional[Dict[str, Any]]): Configuration options for the specific
            storage backend (e.g., authentication details). Defaults to an empty dictionary.
        profile_options (Optional[Dict[str, Any]]): Configuration options for rasterio's
            `to_raster` method when writing the block. Defaults to an empty dictionary.
        overwrite (bool): If True, overwrites the existing blob. Defaults to False.

    Returns:
        int: The number of bytes written, or the size of the existing blob if overwrite is False.

    Raises:
        ValueError: If more than one storage path is identified or no valid components
            are found to form a name.

    Example:
        >>> block = xr.DataArray(...)
        >>> write_block(block, "s3://mybucket/data.tif")
        1024  # hypothetical size of the written data in bytes.
    """
    import rioxarray  # noqa: F401

    if profile_options is None:
        profile_options = {}
    if storage_options is None:
        storage_options = {}

    fs, _, paths = fsspec.get_fs_token_paths(blob_name, storage_options=storage_options)

    if len(paths) > 1:
        msg = "too many paths"
        raise ValueError(msg, paths)

    path = paths[0]

    # Check if the blob exists and overwrite is False, then return its size.
    if not overwrite:  # noqa : SIM103
        if fs.exists(path):
            logging.info("Blob already exists and overwrite is set to False.")
            return fs.info(path)["size"]

    memfs = fsspec.filesystem("memory")

    with memfs.open("data", "wb") as buffer:
        block.squeeze().rio.to_raster(buffer, **profile_options)
        buffer.seek(0)

        if fs.protocol == "file":
            nbytes = len(buffer.getvalue())
        else:
            nbytes = buffer.getbuffer().nbytes

        # use the open method from fsspec to write the buffer
        with fs.open(path, "wb") as f:
            f.write(buffer.getvalue())

    return nbytes


def write_table(
    df: pd.DataFrame | gpd.GeoDataFrame,
    blob_name: str,
    storage_options: dict[str, Any] | None = None,
    overwrite: bool = True,
) -> int | None:
    """
    Write a pandas or geopandas DataFrame to a specified cloud storage location.

    Args:
        df (pd.DataFrame | gpd.GeoDataFrame): The DataFrame to be written to cloud storage.
        blob_name (str): The target storage path, including the blob's name.
        storage_options (Optional[Dict[str, Any]]): Configuration options for the specific
            storage backend (e.g., authentication details). Defaults to an empty dictionary.
        overwrite (bool): If True, overwrites the existing blob. Defaults to False.

    Returns:
        int: The number of bytes written, or the size of the existing blob if overwrite is False.

    Raises:
        ValueError: If more than one storage path is identified.

    Example:
        >>> df = pd.DataFrame(...)
        >>> write_table(df, "s3://mybucket/data.parquet")
    """
    if df.empty:
        return 0

    if storage_options is None:
        storage_options = {}

    fs, _, paths = fsspec.get_fs_token_paths(blob_name, storage_options=storage_options)

    if len(paths) > 1:
        msg = "Too many paths identified"
        raise ValueError(msg, paths)

    path = paths[0]

    # Check if the blob exists and overwrite is False, then return its size.
    if not overwrite and fs.exists(path):
        logging.info("Blob already exists and overwrite is set to False.")
        return fs.info(path)["size"]

    # TODO: return size of table when written to storage
    # Write DataFrame to a buffer
    with fsspec.open(blob_name, "wb", **storage_options) as f:
        df.to_parquet(f)
